__all__ = ("science_radar_batch",)

import astropy.units as u
import healpy as hp
import numpy as np

import rubin_sim.maf as maf
import rubin_sim.maf.maps as maps
import rubin_sim.maf.metric_bundles as mb
import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.plots as plots
import rubin_sim.maf.slicers as slicers

from .common import extended_summary, filter_list, lightcurve_summary, standard_summary


def science_radar_batch(
    runName="run name",
    nside=64,
    benchmarkArea=18000,
    benchmarkNvisits=825,
    minNvisits=750,
    long_microlensing=True,
    srd_only=False,
    mjd0=None,
):
    """A batch of metrics for looking at survey performance relative to the
    SRD and the main science drivers of LSST.

    Parameters
    ----------
    runName : `str`, optional
        The name of the opsim run.
    nside : `int`, optional
        The nside to use for the HealpixSlicer (where relevant).
    benchmarkArea : `float`, optional
        The benchmark value to use for fO_Area comparison.
    benchmarkNvisits : `float`, optional
        The benchmark value to use for fO_Nvisits comparisons.
    minNvisits : `float`, optional
        The value to use to establish the area covered by at least minNvis.
    long_microlensing : `bool`, optional
        Add the longer running microlensing metrics to the batch
        (at a subset of crossing times only)
    srd_only : `bool` , optional
        Only return the SRD metrics
    mjd0 : `float`, optional
        Set the start time for the survey, for metrics which need
        this information.

    Returns
    -------
    metric_bundleDict : `dict` of `maf.MetricBundle`
    """

    bundleList = []
    # Get some standard per-filter coloring and sql constraints
    filterlist, colors, filterorders, filtersqls, filterinfo_label = filter_list(all=False)
    (
        allfilterlist,
        allcolors,
        allfilterorders,
        allfiltersqls,
        allfilterinfo_label,
    ) = filter_list(all=True)

    standardStats = standard_summary(with_count=False)

    # This is the default slicer for most purposes in this batch.
    # Note that the cache is on - if the metric requires a dust map,
    # this is not the right slicer to use.
    healpixslicer = slicers.HealpixSlicer(nside=nside, use_cache=True)
    subsetPlots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]

    #########################
    #########################
    # SRD, DM, etc
    #########################
    #########################
    # fO metric
    displayDict = {"group": "SRD", "subgroup": "FO metrics", "order": 0}

    # Configure the count metric which is what is used for f0 slicer.
    metric = metrics.CountExplimMetric(col="observationStartMJD", metric_name="fO")
    plotDict = {
        "xlabel": "Number of Visits",
        "Asky": benchmarkArea,
        "Nvisit": benchmarkNvisits,
        "x_min": 0,
        "x_max": 1500,
    }
    summaryMetrics = [
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name="fOArea",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=True,
            metric_name="fOArea/benchmark",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FONv(
            nside=nside,
            norm=False,
            metric_name="fONv",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FONv(
            nside=nside,
            norm=True,
            metric_name="fONv/benchmark",
            asky=benchmarkArea,
            n_visit=benchmarkNvisits,
        ),
        metrics.FOArea(
            nside=nside,
            norm=False,
            metric_name=f"fOArea_{minNvisits}",
            asky=benchmarkArea,
            n_visit=minNvisits,
        ),
    ]
    caption = "The FO metric evaluates the overall efficiency of observing. "
    caption += (
        "foNv: out of %.2f sq degrees, the area receives at least X and a median of Y visits "
        "(out of %d, if compared to benchmark). " % (benchmarkArea, benchmarkNvisits)
    )
    caption += (
        "fOArea: this many sq deg (out of %.2f sq deg if compared "
        "to benchmark) receives at least %d visits. " % (benchmarkArea, benchmarkNvisits)
    )
    displayDict["caption"] = caption
    bundle = mb.MetricBundle(
        metric,
        healpixslicer,
        "",
        plot_dict=plotDict,
        display_dict=displayDict,
        summary_metrics=summaryMetrics,
        plot_funcs=[plots.FOPlot()],
    )
    bundleList.append(bundle)

    # Single visit depth distribution
    displayDict["subgroup"] = "Visit Depths"
    # Histogram values over all and per filter.

    value = "fiveSigmaDepth"
    for f in filterlist:
        displayDict["caption"] = "Histogram of %s" % (value)
        displayDict["caption"] += " for %s." % (filterinfo_label[f])
        displayDict["order"] = filterorders[f]
        m = metrics.CountMetric(value, metric_name="%s Histogram" % (value))
        slicer = slicers.OneDSlicer(slice_col_name=value)
        bundle = mb.MetricBundle(
            m,
            slicer,
            filtersqls[f],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    displayDict["caption"] = ""
    for f in filterlist:
        slicer = maf.UniSlicer()
        metric = maf.MedianMetric(col="fiveSigmaDepth")
        bundle = mb.MetricBundle(
            metric,
            slicer,
            filtersqls[f],
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    ##############
    # Astrometry
    ###############

    rmags_para = [22.4, 24.0]
    rmags_pm = [20.5, 24.0]

    # Set up parallax/dcr stackers.
    parallaxStacker = maf.ParallaxFactorStacker()
    dcrStacker = maf.DcrStacker()

    # Set up parallax metrics.
    displayDict["subgroup"] = "Parallax"
    displayDict["order"] += 1
    # Expected error on parallax at 10 AU.
    plotmaxVals = (2.0, 15.0)
    good_parallax_limit = 11.5
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.median,
            decreasing=False,
            metric_name="Median Parallax Uncert (18k)",
        ),
        metrics.AreaThresholdMetric(
            upper_threshold=good_parallax_limit,
            metric_name="Area better than %.1f mas uncertainty" % good_parallax_limit,
        ),
    ]
    summary.append(metrics.PercentileMetric(percentile=95, metric_name="95th Percentile Parallax Uncert"))
    summary.extend(standard_summary())
    for rmag, plotmax in zip(rmags_para, plotmaxVals):
        plotDict = {"x_min": 0, "x_max": plotmax, "color_min": 0, "color_max": plotmax}
        metric = metrics.ParallaxMetric(
            metric_name="Parallax Uncert @ %.1f" % (rmag),
            rmag=rmag,
            normalize=False,
        )
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # Parallax normalized to 'best possible'.
    # This separates the effect of cadence from depth.
    for rmag in rmags_para:
        metric = metrics.ParallaxMetric(
            metric_name="Normalized Parallax Uncert @ %.1f" % (rmag),
            rmag=rmag,
            normalize=True,
        )
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Parallax factor coverage.
    for rmag in rmags_para:
        metric = metrics.ParallaxCoverageMetric(metric_name="Parallax Coverage @ %.1f" % (rmag), rmag=rmag)
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            stacker_list=[parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Parallax problems can be caused by HA and DCR degeneracies.
    # Check their correlation.
    for rmag in rmags_para:
        metric = metrics.ParallaxDcrDegenMetric(
            metric_name="Parallax-DCR degeneracy @ %.1f" % (rmag), rmag=rmag
        )
        caption = "Correlation between parallax offset magnitude and hour angle for a r=%.1f star." % (rmag)
        caption += " (0 is good, near -1 or 1 is bad)."
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            stacker_list=[dcrStacker, parallaxStacker],
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # Proper Motion metrics.
    displayDict["subgroup"] = "Proper Motion"
    displayDict["order"] = 0
    # Proper motion errors.
    plotmaxVals = (1.0, 5.0)
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.median,
            decreasing=False,
            metric_name="Median Proper Motion Uncert (18k)",
        )
    ]
    summary.append(metrics.PercentileMetric(metric_name="95th Percentile Proper Motion Uncert"))
    summary.extend(standard_summary())
    for rmag, plotmax in zip(rmags_pm, plotmaxVals):
        plotDict = {"x_min": 0, "x_max": plotmax, "color_min": 0, "color_max": plotmax}
        metric = metrics.ProperMotionMetric(
            metric_name="Proper Motion Uncert @ %.1f" % rmag,
            rmag=rmag,
            normalize=False,
        )
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            display_dict=displayDict,
            plot_dict=plotDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1
    # Normalized proper motion.
    for rmag in rmags_pm:
        metric = metrics.ProperMotionMetric(
            metric_name="Normalized Proper Motion Uncert @ %.1f" % rmag,
            rmag=rmag,
            normalize=True,
        )
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            "",
            display_dict=displayDict,
            summary_metrics=standard_summary(),
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)
        displayDict["order"] += 1

    # DCR precision metric
    displayDict["subgroup"] = "DCR"
    displayDict["order"] = 0
    plotDict = {"caption": "Precision of DCR slope.", "percentile_clip": 95}
    metric = metrics.DcrPrecisionMetric()
    bundle = mb.MetricBundle(
        metric,
        healpixslicer,
        "",
        plot_dict=plotDict,
        plot_funcs=subsetPlots,
        display_dict=displayDict,
        summary_metrics=standard_summary(with_count=False),
    )
    bundleList.append(bundle)

    # Rapid Revisit
    displayDict["subgroup"] = "Rapid Revisits"
    displayDict["order"] = 0
    # Calculate the actual number of revisits within 30 minutes.
    dTmax = 30  # time in minutes
    m2 = metrics.NRevisitsMetric(d_t=dTmax, normed=False, metric_name="NumberOfQuickRevisits")
    plotDict = {"color_min": 400, "color_max": 2000, "x_min": 400, "x_max": 2000}
    caption = "Number of consecutive visits with return times faster than %.1f minutes, " % (dTmax)
    caption += "in any filter. "
    displayDict["caption"] = caption
    bundle = mb.MetricBundle(
        m2,
        healpixslicer,
        "",
        plot_dict=plotDict,
        plot_funcs=subsetPlots,
        display_dict=displayDict,
        summary_metrics=standard_summary(with_count=False),
    )
    bundleList.append(bundle)
    displayDict["order"] += 1

    # Better version of the rapid revisit requirements:
    # require a minimum number of visits between
    # dtMin and dtMax, but also a minimum number of visits
    # between dtMin and dtPair (the typical pair time).
    # 1 means the healpix met the requirements (0 means did not).
    dTmin = 40.0 / 60.0  # (minutes) 40s minimum for rapid revisit range
    dTpairs = 20.0  # minutes (time when pairs should start kicking in)
    dTmax = 30.0  # 30 minute maximum for rapid revisit range
    nOne = 82  # Number of revisits between 40s-30m required
    nTwo = 28  # Number of revisits between 40s - tPairs required.
    pix_area = float(hp.nside2pixarea(nside, degrees=True))
    scale = pix_area * hp.nside2npix(nside)
    m1 = metrics.RapidRevisitMetric(
        metric_name="RapidRevisits",
        d_tmin=dTmin / 60.0 / 60.0 / 24.0,
        d_tpairs=dTpairs / 60.0 / 24.0,
        d_tmax=dTmax / 60.0 / 24.0,
        min_n1=nOne,
        min_n2=nTwo,
    )
    plotDict = {
        "x_min": 0,
        "x_max": 1,
        "color_min": 0,
        "color_max": 1,
        "log_scale": False,
    }
    cutoff1 = 0.9
    summaryStats = [metrics.FracAboveMetric(cutoff=cutoff1, scale=scale, metric_name="Area (sq deg)")]
    caption = "Rapid Revisit: area that receives at least %d visits between %.3f and %.1f minutes, " % (
        nOne,
        dTmin,
        dTmax,
    )
    caption += "with at least %d of those visits falling between %.3f and %.1f minutes. " % (
        nTwo,
        dTmin,
        dTpairs,
    )
    caption += (
        'Summary statistic "Area" indicates the area on the sky which meets this requirement.'
        " (SRD design specification is 2000 sq deg)."
    )
    displayDict["caption"] = caption
    displayDict["order"] = 0
    bundle = mb.MetricBundle(
        m1,
        healpixslicer,
        "",
        plot_dict=plotDict,
        plot_funcs=subsetPlots,
        display_dict=displayDict,
        summary_metrics=summaryStats,
    )
    bundleList.append(bundle)

    # For SRD batch only, return here.
    if srd_only:
        for b in bundleList:
            b.set_run_name(runName)
        bundleDict = mb.make_bundles_dict_from_list(bundleList)

        return bundleDict

    # Year Coverage
    displayDict["subgroup"] = "Year Coverage"
    metric = metrics.YearCoverageMetric()
    for f in filterlist:
        plotDict = {"color_min": 7, "color_max": 10, "color": colors[f]}
        displayDict["caption"] = f"Number of years of coverage in {f} band."
        displayDict["order"] = filterorders[f]
        summary = [
            metrics.AreaSummaryMetric(
                area=18000,
                reduce_func=np.mean,
                decreasing=True,
                metric_name="N Years (18k) %s" % f,
            )
        ]
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            filtersqls[f],
            plot_dict=plotDict,
            info_label=filterinfo_label[f],
            display_dict=displayDict,
            summary_metrics=summary,
        )
        bundleList.append(bundle)

    #########################
    #########################
    # Galaxies
    #########################
    #########################

    # Run this per filter, to look at variations in counts
    # of galaxies in blue bands?
    displayDict = {
        "group": "Galaxies",
        "subgroup": "Galaxy Counts",
        "order": 0,
        "caption": None,
    }
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.sum,
            decreasing=True,
            metric_name="N Galaxies (18k)",
        )
    ]
    summary.append(metrics.SumMetric(metric_name="N Galaxies (all)"))
    # make sure slicer has cache off
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    for f in filterlist:
        plotDict = {"percentile_clip": 95.0, "n_ticks": 5, "color": colors[f]}
        metric = maf.GalaxyCountsMetricExtended(filter_band=f, redshift_bin="all", nside=nside)
        displayDict[
            "caption"
        ] = f"Number of galaxies across the sky, in {f} band. Generally, full survey footprint."
        displayDict["order"] = filterorders[f]
        bundle = mb.MetricBundle(
            metric,
            slicer,
            filtersqls[f],
            plot_dict=plotDict,
            display_dict=displayDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)

    displayDict["subgroup"] = "Surface Brightness"
    summary = [metrics.MedianMetric()]
    for filtername in "ugrizy":
        displayDict["caption"] = "Surface brightness limit in %s, no extinction applied." % filtername
        displayDict["order"] = filterorders[f]
        sql = 'filter="%s"' % filtername
        metric = metrics.SurfaceBrightLimitMetric()
        bundle = mb.MetricBundle(
            metric,
            healpixslicer,
            sql,
            display_dict=displayDict,
            summary_metrics=summary,
            plot_funcs=subsetPlots,
        )
        bundleList.append(bundle)

    #########################
    #########################
    # Cosmology
    #########################
    #########################

    bandpass = "i"
    nfilters_needed = 6
    lim_ebv = 0.2
    offset = 0.1
    mag_cuts = {
        1: 24.75 - offset,
        2: 25.12 - offset,
        3: 25.35 - offset,
        4: 25.5 - offset,
        5: 25.62 - offset,
        6: 25.72 - offset,
        7: 25.8 - offset,
        8: 25.87 - offset,
        9: 25.94 - offset,
        10: 26.0 - offset,
    }
    yrs = list(mag_cuts.keys())
    maxYr = max(yrs)

    displayDict = {"group": "Cosmology"}
    subgroupCount = 1

    displayDict["subgroup"] = f"{subgroupCount}: Static Science"
    ## Static Science
    # Calculate the static science metrics - effective survey area,
    # mean/median coadded depth, stdev of coadded depth and
    # the 3x2ptFoM emulator.

    dustmap = maps.DustMap(nside=nside, interp=False)
    pix_area = hp.nside2pixarea(nside, degrees=True)
    summaryMetrics = [
        metrics.MeanMetric(),
        metrics.MedianMetric(),
        metrics.RmsMetric(),
        metrics.CountRatioMetric(norm_val=1 / pix_area, metric_name="Effective Area (deg)"),
    ]
    displayDict["order"] = 0
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    for yr_cut in yrs:
        ptsrc_lim_mag_i_band = mag_cuts[yr_cut]
        sqlconstraint = "night <= %s" % (yr_cut * 365.25 + 0.5)
        sqlconstraint += ' and note not like "DD%"'
        info_label = f"{bandpass} band non-DD year {yr_cut}"
        ThreebyTwoSummary_simple = metrics.StaticProbesFoMEmulatorMetricSimple(
            nside=nside, year=yr_cut, metric_name="3x2ptFoM_simple"
        )
        ThreebyTwoSummary = maf.StaticProbesFoMEmulatorMetric(nside=nside, metric_name="3x2ptFoM")

        m = metrics.ExgalM5WithCuts(
            lsst_filter=bandpass,
            n_filters=nfilters_needed,
            extinction_cut=lim_ebv,
            depth_cut=ptsrc_lim_mag_i_band,
        )
        caption = (
            f"Cosmology/Static science metrics are based on evaluating the region "
            f"of the sky that meets the requirements (in year {yr_cut} of coverage"
            f"in all {nfilters_needed} bands, a lower E(B-V) value than {lim_ebv} "
            f"and at least a coadded depth of {ptsrc_lim_mag_i_band} in {bandpass}. "
            f"From there the effective survey area, coadded depth, standard deviation of "
            f"the depth, and a 3x2pt static science figure of merit emulator are "
            f"calculated using the dust-extinction coadded depth map (over that reduced "
            f"footprint)."
        )
        displayDict["caption"] = caption
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqlconstraint,
            maps_list=[dustmap],
            info_label=info_label,
            summary_metrics=summaryMetrics + [ThreebyTwoSummary, ThreebyTwoSummary_simple],
            display_dict=displayDict,
        )
        displayDict["order"] += 1
        bundleList.append(bundle)

    ## LSS Science
    # The only metric we have from LSS is the NGals metric -
    # which is similar to the GalaxyCountsExtended
    # metric, but evaluated only on the depth/dust cuts footprint.
    subgroupCount += 1
    displayDict["subgroup"] = f"{subgroupCount}: LSS"
    displayDict["order"] = 0
    plotDict = {"n_ticks": 5}
    # Have to include all filters in query to check for filter coverage.
    # Galaxy numbers calculated using 'bandpass' images only though.
    sqlconstraint = 'note not like "DD%"'
    info_label = f"{bandpass} band galaxies non-DD"
    metric = maf.DepthLimitedNumGalMetric(
        nside=nside,
        filter_band=bandpass,
        redshift_bin="all",
        nfilters_needed=nfilters_needed,
        lim_mag_i_ptsrc=mag_cuts[maxYr],
        lim_ebv=lim_ebv,
    )
    summary = [
        metrics.AreaSummaryMetric(
            area=18000,
            reduce_func=np.sum,
            decreasing=True,
            metric_name="N Galaxies (18k)",
        )
    ]
    summary.append(metrics.SumMetric(metric_name="N Galaxies (all)"))
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sqlconstraint,
        plot_dict=plotDict,
        info_label=info_label,
        maps_list=[dustmap],
        display_dict=displayDict,
        summary_metrics=summary,
        plot_funcs=subsetPlots,
    )
    bundleList.append(bundle)

    ## WL metrics
    # Calculates the number of visits per pointing,
    # after removing parts of the footprint due to dust/depth.
    # Count visits in gri bands.
    subgroupCount += 1
    displayDict["subgroup"] = f"{subgroupCount}: WL"
    displayDict["order"] = 0
    sqlconstraint = 'note not like "DD%" and (filter="g" or filter="r" or filter="i")'
    info_label = "gri band non-DD"
    minExpTime = 15
    m = metrics.WeakLensingNvisits(
        lsst_filter=bandpass,
        depth_cut=mag_cuts[maxYr],
        ebvlim=lim_ebv,
        min_exp_time=minExpTime,
        metric_name="WeakLensingNvisits",
    )
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    displayDict["caption"] = (
        f"The number of visits per pointing, over a similarly reduced footprint as "
        f"described above for the 3x2pt FOM, but allowing areas of sky with "
        f"fewer than {nfilters_needed} filters. "
        f"A cutoff of {minExpTime} removes very short visits."
    )
    bundle = mb.MetricBundle(
        m,
        slicer,
        sqlconstraint,
        maps_list=[dustmap],
        info_label=info_label,
        summary_metrics=standardStats,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Do the weak lensing per year
    for year in np.arange(1, 10):
        displayDict["order"] = year
        sqlconstraint = (
            'note not like "DD%"'
            + ' and (filter="g" or filter="r" or filter="i") and night < %i' % (year * 365.25)
        )
        m = metrics.WeakLensingNvisits(
            lsst_filter=bandpass,
            depth_cut=mag_cuts[year],
            ebvlim=lim_ebv,
            min_exp_time=minExpTime,
            metric_name="WeakLensingNvisits_gri_year%i" % year,
        )
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqlconstraint,
            maps_list=[dustmap],
            info_label=info_label,
            summary_metrics=standardStats,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

        m = metrics.RIZDetectionCoaddExposureTime(
            det_bands=["g", "r", "i"], metric_name="gri_exposure_time_year%i" % year
        )
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqlconstraint,
            maps_list=[dustmap],
            info_label=info_label,
            summary_metrics=standardStats,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

        sqlconstraint = (
            'note not like "DD%"'
            + ' and (filter="r" or filter="i" or filter="z") and night < %i' % (year * 365.25)
        )
        m = metrics.WeakLensingNvisits(
            lsst_filter=bandpass,
            depth_cut=mag_cuts[year],
            ebvlim=lim_ebv,
            min_exp_time=minExpTime,
            metric_name="WeakLensingNvisits_riz_year%i" % year,
        )
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqlconstraint,
            maps_list=[dustmap],
            info_label=info_label,
            summary_metrics=standardStats,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

        m = metrics.RIZDetectionCoaddExposureTime(
            det_bands=["g", "r", "i"], metric_name="riz_exposure_time_year%i" % year
        )
        bundle = mb.MetricBundle(
            m,
            slicer,
            sqlconstraint,
            maps_list=[dustmap],
            info_label=info_label,
            summary_metrics=standardStats,
            display_dict=displayDict,
        )
        bundleList.append(bundle)

    subgroupCount += 1
    displayDict["subgroup"] = f"{subgroupCount}: Camera Rotator"
    metric1 = metrics.KuiperMetric("rotSkyPos")
    metric2 = metrics.KuiperMetric("rotTelPos")
    caption_root = "Kuiper statistic (0 is uniform, 1 is delta function) of the "
    for f in filterlist:
        for m in [metric1, metric2]:
            plotDict = {"color": colors[f]}
            displayDict["order"] = filterorders[f]
            displayDict["caption"] = caption_root + f"{m.colname} for visits in {f} band."
            bundleList.append(
                mb.MetricBundle(
                    m,
                    healpixslicer,
                    filtersqls[f],
                    plot_dict=plotDict,
                    display_dict=displayDict,
                    summary_metrics=standardStats,
                    plot_funcs=subsetPlots,
                )
            )
    # Kuiper per year in gri and riz
    for year in np.arange(1, 10):
        sqlconstraint = (
            'note not like "DD%"'
            + ' and (filter="g" or filter="r" or filter="i") and night < %i' % (year * 365.25)
        )
        metric1 = metrics.KuiperMetric("rotSkyPos", metric_name="Kuiper_rotSkyPos_gri_year%i" % year)
        metric2 = metrics.KuiperMetric("rotTelPos", metric_name="Kuiper_rotTelPos_gri_year%i" % year)
        for metric in [metric1, metric2]:
            bundleList.append(
                mb.MetricBundle(
                    metric,
                    healpixslicer,
                    sqlconstraint,
                    plot_dict={},
                    display_dict=displayDict,
                    summary_metrics=standardStats,
                    plot_funcs=subsetPlots,
                )
            )

    ##############
    # SNe Ia
    ##############
    displayDict = {
        "group": "Cosmology",
        "subgroup": "5: SNe Ia",
        "order": 0,
        "caption": "Expected discoveries of SNeIa, using the SNNSNMetric.",
    }
    sne_nside = 16
    sn_summary = [
        metrics.MedianMetric(),
        metrics.MeanMetric(),
        metrics.SumMetric(metric_name="Total detected"),
        metrics.CountMetric(metric_name="Total on sky", mask_val=0),
    ]
    snslicer = slicers.HealpixSlicer(nside=sne_nside, use_cache=False)
    metric = metrics.SNNSNMetric(
        n_bef=3,
        n_aft=8,
        coadd_night=True,
        add_dust=False,
        hard_dust_cut=0.25,
        zmin=0.2,
        zmax=0.5,
        z_step=0.03,
        daymax_step=3.0,
        zlim_coeff=0.95,
        gamma_name="gamma_WFD.hdf5",
        verbose=False,
    )
    plotDict = {"percentile_clip": 95, "n_ticks": 5}
    # Run without DDF observations
    bundle = mb.MetricBundle(
        metric,
        snslicer,
        "note not like '%DD%'",
        plot_dict=plotDict,
        display_dict=displayDict,
        info_label="DDF excluded",
        summary_metrics=sn_summary,
        plot_funcs=subsetPlots,
    )

    bundleList.append(bundle)

    #########################
    #########################
    # AGN
    #########################
    #########################

    # AGN structure function error
    agnslicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    dustmap = maps.DustMap(nside=nside)
    displayDict = {"group": "AGN", "order": 0}

    # Calculate the number of expected QSOs, in each band
    for f in filterlist:
        sql = filtersqls[f] + ' and note not like "%DD%"'
        md = filterinfo_label[f] + " and non-DD"
        summaryMetrics = [metrics.SumMetric(metric_name="Total QSO")]
        zmin = 0.3
        m = metrics.QSONumberCountsMetric(
            f,
            units="mag",
            extinction_cut=1.0,
            qlf_module="Shen20",
            qlf_model="A",
            sed_model="Richards06",
            zmin=zmin,
            zmax=None,
        )
        displayDict["subgroup"] = "nQSO"
        displayDict["caption"] = (
            "The expected number of QSOs in regions of low dust extinction,"
            f"based on detection in {f} bandpass."
        )
        displayDict["order"] = filterorders[f]
        bundleList.append(
            mb.MetricBundle(
                m,
                agnslicer,
                constraint=sql,
                info_label=md,
                run_name=runName,
                maps_list=[dustmap],
                summary_metrics=summaryMetrics,
                display_dict=displayDict,
            )
        )

    # Calculate the expected AGN structure function error
    # These agn test magnitude values are determined by
    # looking at the baseline median m5 depths
    # For v1.7.1 these values are:
    agn_m5 = {"u": 22.89, "g": 23.94, "r": 23.5, "i": 22.93, "z": 22.28, "y": 21.5}
    # And the expected medians SF error at those values is about 0.04
    summaryMetrics = extended_summary()
    summaryMetrics += [metrics.AreaThresholdMetric(upper_threshold=0.03, metric_name="AreaThreshold_0.03")]
    summaryMetrics += [metrics.AreaThresholdMetric(upper_threshold=0.06, metric_name="AreaThreshold_0.06")]
    for f in filterlist:
        m = metrics.SFUncertMetric(
            mag=agn_m5[f],
            bins=np.logspace(0, np.log10(3650), 16),
            metric_name="AGN SF_uncert",
        )
        plotDict = {"color": colors[f], "color_min": 0, "color_max": 0.2}
        displayDict["order"] = filterorders[f]
        displayDict["subgroup"] = "SFUncert"
        displayDict["caption"] = (
            "Expected AGN structure function uncertainties, based on observations in "
            f"{f} band, for an AGN of magnitude {agn_m5[f]:.2f}"
        )
        bundleList.append(
            mb.MetricBundle(
                m,
                agnslicer,
                constraint=filtersqls[f],
                info_label=filterinfo_label[f],
                run_name=runName,
                maps_list=[dustmap],
                plot_dict=plotDict,
                summary_metrics=summaryMetrics,
                display_dict=displayDict,
            )
        )

    # Run the TimeLag for each filter *and* all filters
    nquist_threshold = 2.2
    lag = 100
    summaryMetrics = extended_summary()
    summaryMetrics += [metrics.AreaThresholdMetric(lower_threshold=nquist_threshold)]
    m = metrics.AgnTimeLagMetric(threshold=nquist_threshold, lag=lag)
    for f in allfilterlist:
        plotDict = {
            "color": allcolors[f],
            "color_min": 0,
            "color_max": 5,
            "percentile_clip": 95,
        }
        displayDict["order"] = allfilterorders[f]
        displayDict["subgroup"] = "Time Lags"
        displayDict["caption"] = (
            f"Comparion of the time between visits compared to a defined sampling gap ({lag} days) in "
            f"{f} band."
        )
        bundleList.append(
            mb.MetricBundle(
                m,
                agnslicer,
                constraint=allfiltersqls[f],
                info_label=allfilterinfo_label[f],
                run_name=runName,
                maps_list=[dustmap],
                plot_dict=plotDict,
                summary_metrics=summaryMetrics,
                display_dict=displayDict,
            )
        )

    # Run the TimeLag for each filter *and* all filters but for 5 days
    nquist_threshold = 2.2
    lag = 5
    summaryMetrics = extended_summary()
    summaryMetrics += [metrics.AreaThresholdMetric(lower_threshold=nquist_threshold)]
    m = metrics.AgnTimeLagMetric(threshold=nquist_threshold, lag=lag)
    for f in allfilterlist:
        plotDict = {
            "color": allcolors[f],
            "color_min": 0,
            "color_max": 5,
            "percentile_clip": 95,
        }
        displayDict["order"] = allfilterorders[f]
        displayDict["subgroup"] = "Time Lags"
        displayDict["caption"] = (
            f"Comparion of the time between visits compared to a defined sampling gap ({lag} days) in "
            f"{f} band."
        )
        bundleList.append(
            mb.MetricBundle(
                m,
                agnslicer,
                constraint=allfiltersqls[f],
                info_label=allfilterinfo_label[f],
                run_name=runName,
                maps_list=[dustmap],
                plot_dict=plotDict,
                summary_metrics=summaryMetrics,
                display_dict=displayDict,
            )
        )

    #########################
    #########################
    # Strong Lensing
    #########################
    #########################

    # TDC metric
    # Calculate a subset of DESC WFD-related metrics.
    if nside > 64:
        nside_tdc = 64
    else:
        nside_tdc = nside
    displayDict = {"group": "Strong Lensing"}
    displayDict["subgroup"] = "Lens Time Delay"

    tdc_plots = [plots.HealpixSkyMap(), plots.HealpixHistogram()]
    plotDict = {"x_min": 0.01, "color_min": 0.01, "percentile_clip": 70, "n_ticks": 5}
    tdc_summary = [metrics.MeanMetric(), metrics.MedianMetric(), metrics.RmsMetric()]
    # Ideally need a way to do better on calculating the summary metrics
    # for the high accuracy area.
    slicer = slicers.HealpixSlicer(nside=nside_tdc, use_cache=False)
    tdcMetric = metrics.TdcMetric(metric_name="TDC")
    dustmap = maps.DustMap(nside=nside_tdc, interp=False)
    bundle = mb.MetricBundle(
        tdcMetric,
        slicer,
        constraint="",
        display_dict=displayDict,
        plot_dict=plotDict,
        plot_funcs=tdc_plots,
        maps_list=[dustmap],
        summary_metrics=tdc_summary,
    )
    bundleList.append(bundle)

    # Strongly lensed SNe
    displayDict["group"] = "Strong Lensing"
    displayDict["subgroup"] = "SLSN"
    displayDict["caption"] = "Strongly Lensed SNe, evaluated with the addition of galactic dust extinction."
    metric = metrics.SNSLMetric()
    slicer = slicers.HealpixSlicer(nside=nside_tdc, use_cache=False)
    plotDict = {}
    bundle = mb.MetricBundle(
        metric,
        slicer,
        "",
        run_name=runName,
        plot_dict=plotDict,
        summary_metrics=metrics.SumMetric(metric_name="Total detected"),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    #########################
    #########################
    # Variables and Transients
    #########################
    #########################

    # Periodic Stars
    displayDict = {"group": "Variables/Transients", "order": 0}

    # PeriodicStarModulation metric
    # colors for c type RRLyrae
    displayDict["subgroup"] = "Periodic Star Modulation"
    I_rrc_lmc = 18.9
    V_rrc_lmc = 19.2
    Vi = V_rrc_lmc - (2.742 * 0.08) - 18.5
    Ii = I_rrc_lmc - (1.505 * 0.08) - 18.5
    ii_rrc = Ii + 0.386 * 0.013 + 0.397  # 0.013 = (i-z)_0
    gi_rrc = ii_rrc + 1.481 * (Vi - Ii) - 0.536
    ri_rrc = (1 / 0.565) * (Vi - 0.435 * gi_rrc + 0.016)
    ui_rrc = gi_rrc + 0.575
    zi_rrc = ii_rrc - 0.013
    yi_rrc = zi_rrc
    rrc = np.array([ui_rrc, gi_rrc, ri_rrc, ii_rrc, zi_rrc, yi_rrc])
    time_intervals = (15, 30)
    distMod = (18, 19, 20, 21)
    summaryStats = [metrics.MeanMetric(), metrics.MedianMetric(), metrics.MaxMetric()]
    slicer = slicers.HealpixSlicer(nside=8)
    for time_interval in time_intervals:
        for dM in distMod:
            displayDict["caption"] = (
                "Periodic star modulation metric, evaluates the likelihood of "
                "measuring variation in an RRLyrae periodic variable. "
                "Evaluated based on the full LSST survey data. "
                f"Searching time interval of {time_interval} and distance modulus {dM}."
            )
            displayDict["order"] += 1
            m = maf.PeriodicStarModulationMetric(
                period=0.3,
                amplitude=0.3,
                random_phase=True,
                time_interval=time_interval,
                n_monte=100,
                period_tol=0.002,
                amp_tol=0.01,
                means=rrc + dM,
                mag_tol=0.01,
                n_bands=3,
            )
            # Run this on year 1-2.
            bundle = mb.MetricBundle(
                m,
                slicer,
                "night < 365.25*2",
                display_dict=displayDict,
                run_name=runName,
                summary_metrics=summaryStats,
                info_label=f"dm {dM} interval {time_interval} RRc Year 1-2",
            )
            bundleList.append(bundle)

    # PulsatingStarRecovery metric (to be added; Marcella)

    # our periodic star metrics
    displayDict["subgroup"] = "Periodic Stars"
    displayDict["order"] = 0
    for period in [0.5, 1, 2]:
        for magnitude in [21.0, 24.0]:
            amplitudes = [0.05, 0.1, 1.0]
            periods = [period] * len(amplitudes)
            starMags = [magnitude] * len(amplitudes)

            plotDict = {
                "n_ticks": 3,
                "color_min": 0,
                "color_max": 3,
                "x_min": 0,
                "x_max": 3,
            }
            info_label = "P_%.1f_Mag_%.0f_Amp_0.05-0.1-1" % (period, magnitude)
            sql = ""
            displayDict["caption"] = (
                "Metric evaluates if a periodic signal of period %.1f days could "
                "be detected for an r=%i star. A variety of amplitudes of periodicity "
                "are tested: [1, 0.1, and 0.05] mag amplitudes, which correspond to "
                "metric values of [1, 2, or 3]. " % (period, magnitude)
            )
            metric = metrics.PeriodicDetectMetric(
                periods=periods,
                star_mags=starMags,
                amplitudes=amplitudes,
                metric_name="PeriodicDetect",
            )
            bundle = mb.MetricBundle(
                metric,
                healpixslicer,
                sql,
                info_label=info_label,
                display_dict=displayDict,
                plot_dict=plotDict,
                plot_funcs=subsetPlots,
                summary_metrics=standardStats,
            )
            bundleList.append(bundle)
            displayDict["order"] += 1

    # our periodic star metrics - first two years only
    displayDict["order"] = 1
    for period in [0.5, 1, 2]:
        for magnitude in [21.0, 24.0]:
            amplitudes = [0.05, 0.1, 1.0]
            periods = [period] * len(amplitudes)
            starMags = [magnitude] * len(amplitudes)

            plotDict = {
                "n_ticks": 3,
                "color_min": 0,
                "color_max": 3,
                "x_min": 0,
                "x_max": 3,
            }
            info_label = "P_%.1f_Mag_%.0f_Amp_0.05-0.1-1 Yr 2" % (period, magnitude)
            sql = "night < 365.5*2"
            displayDict["caption"] = (
                "Metric evaluates if a periodic signal of period %.1f days could "
                "be detected for an r=%i star, within the first two years. "
                "A variety of amplitudes of periodicity "
                "are tested: [1, 0.1, and 0.05] mag amplitudes, which correspond to "
                "metric values of [1, 2, or 3]. " % (period, magnitude)
            )
            metric = metrics.PeriodicDetectMetric(
                periods=periods,
                star_mags=starMags,
                amplitudes=amplitudes,
                metric_name="PeriodicDetect",
            )
            bundle = mb.MetricBundle(
                metric,
                healpixslicer,
                sql,
                info_label=info_label,
                display_dict=displayDict,
                plot_dict=plotDict,
                plot_funcs=subsetPlots,
                summary_metrics=standardStats,
            )
            bundleList.append(bundle)
            displayDict["order"] += 1

    # Tidal Disruption Events
    displayDict["subgroup"] = "TDE"
    displayDict["caption"] = "TDE lightcurves that could be identified"
    displayDict["order"] = 0
    metric = maf.TdePopMetric(mjd0=mjd0)
    tdeslicer = maf.generate_tde_pop_slicer()
    bundle = mb.MetricBundle(
        metric,
        tdeslicer,
        "",
        run_name=runName,
        summary_metrics=lightcurve_summary(),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    displayDict["caption"] = "TDE lightcurves quality"
    metric = maf.TdePopMetricQuality(metric_name="TDE_Quality")
    bundle = mb.MetricBundle(
        metric,
        tdeslicer,
        "",
        run_name=runName,
        summary_metrics=lightcurve_summary(),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Microlensing events

    displayDict["subgroup"] = "Microlensing"
    plotDict = {"nside": 128}

    n_events = 10000
    # Let's evaluate a variety of crossing times
    crossing_times = [
        [1, 5],
        [5, 10],
        [10, 20],
        [20, 30],
        [30, 60],
        [60, 90],
        [100, 200],
        [200, 500],
        [500, 1000],
    ]
    metric = maf.MicrolensingMetric()
    summaryMetrics = maf.batches.lightcurve_summary()
    order = 0
    for crossing in crossing_times:
        displayDict["caption"] = "Microlensing events with crossing times between %i to %i days." % (
            crossing[0],
            crossing[1],
        )
        displayDict["order"] = order
        order += 1
        slicer = maf.generate_microlensing_slicer(
            min_crossing_time=crossing[0],
            max_crossing_time=crossing[1],
            n_events=n_events,
        )
        bundleList.append(
            maf.MetricBundle(
                metric,
                slicer,
                None,
                run_name=runName,
                summary_metrics=summaryMetrics,
                info_label=f"tE {crossing[0]}_{crossing[1]} days",
                display_dict=displayDict,
                plot_dict=plotDict,
                plot_funcs=[plots.HealpixSkyMap()],
            )
        )

    if long_microlensing:
        n_events = 10000
        # Let's evaluate a subset of the crossing times for these
        crossing_times = [
            [10, 20],
            [20, 30],
            [30, 60],
            [200, 500],
        ]
        metric_Npts = maf.MicrolensingMetric(metric_calc="Npts")
        summaryMetrics = maf.batches.microlensing_summary(metric_type="Npts")
        order = 0
        for crossing in crossing_times:
            slicer = maf.generate_microlensing_slicer(
                min_crossing_time=crossing[0],
                max_crossing_time=crossing[1],
                n_events=n_events,
            )
            displayDict["caption"] = "Microlensing events with crossing times between %i to %i days." % (
                crossing[0],
                crossing[1],
            )
            displayDict["order"] = order
            order += 1
            bundleList.append(
                maf.MetricBundle(
                    metric_Npts,
                    slicer,
                    None,
                    run_name=runName,
                    summary_metrics=summaryMetrics,
                    info_label=f"tE {crossing[0]}_{crossing[1]} days",
                    display_dict=displayDict,
                    plot_dict=plotDict,
                    plot_funcs=[],
                )
            )

        metric_Fisher = maf.MicrolensingMetric(metric_calc="Fisher")
        summaryMetrics = maf.batches.microlensing_summary(metric_type="Fisher")
        order = 0
        for crossing in crossing_times:
            displayDict["caption"] = "Microlensing events with crossing times between %i to %i days." % (
                crossing[0],
                crossing[1],
            )
            displayDict["order"] = order
            order += 1
            slicer = maf.generate_microlensing_slicer(
                min_crossing_time=crossing[0],
                max_crossing_time=crossing[1],
                n_events=n_events,
            )
            bundleList.append(
                maf.MetricBundle(
                    metric_Fisher,
                    slicer,
                    None,
                    run_name=runName,
                    summary_metrics=summaryMetrics,
                    info_label=f"tE {crossing[0]}_{crossing[1]} days",
                    display_dict=displayDict,
                    plot_dict=plotDict,
                    plot_funcs=[],
                )
            )

    # Kilonovae metric
    displayDict["group"] = "Variables/Transients"
    displayDict["subgroup"] = "KNe"
    n_events = 500000
    caption = f"KNe metric, injecting {n_events} lightcurves over the entire sky, GW170817-like only."
    caption += " Ignoring DDF observations."
    displayDict["caption"] = caption
    displayDict["order"] = 0
    # Kilonova parameters
    inj_params_list = [
        {"mej_dyn": 0.005, "mej_wind": 0.050, "phi": 30, "theta": 25.8},
    ]
    filename = maf.get_kne_filename(inj_params_list)
    kneslicer = maf.generate_kn_pop_slicer(n_events=n_events, n_files=len(filename), d_min=10, d_max=600)
    metric = maf.KNePopMetric(
        output_lc=False,
        file_list=filename,
        metric_name="KNePopMetric_single",
        mjd0=mjd0,
    )
    bundle = mb.MetricBundle(
        metric,
        kneslicer,
        "note not like 'DD%'",
        run_name=runName,
        info_label="single model",
        summary_metrics=lightcurve_summary(),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    n_events = 500000
    caption = f"KNe metric, injecting {n_events} lightcurves over the entire sky, entire model population."
    caption += " Ignoring DDF observations."
    displayDict["caption"] = caption
    displayDict["order"] = 1
    # Kilonova parameters
    filename = maf.get_kne_filename(None)
    kneslicer_allkne = maf.generate_kn_pop_slicer(
        n_events=n_events, n_files=len(filename), d_min=10, d_max=600
    )
    metric_allkne = maf.KNePopMetric(output_lc=False, file_list=filename, metric_name="KNePopMetric_all")
    bundle = mb.MetricBundle(
        metric_allkne,
        kneslicer_allkne,
        "note not like 'DD%'",
        run_name=runName,
        info_label="all models",
        summary_metrics=lightcurve_summary(),
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # General time intervals
    displayDict = {
        "group": "TimeGaps",
        "subgroup": "Time",
        "caption": None,
        "order": 0,
    }

    # Logarithmically spaced gaps from 30s to 5 years
    tMin = 30 / 60 / 60 / 24.0  # 30s
    tMax = 5 * 365.25  # 5 years
    tgaps = np.logspace(np.log10(tMin), np.log10(tMax), 100)

    for f in filterlist:
        m1 = metrics.TgapsMetric(bins=tgaps, all_gaps=False)
        plotDict = {
            "bins": tgaps,
            "xscale": "log",
            "y_min": 0,
            "figsize": (8, 6),
            "ylabel": "Number of observation pairs",
            "xlabel": "Time gap between pairs of visits (days)",
            "color": colors[f],
        }
        plotFuncs = [plots.SummaryHistogram()]
        displayDict["caption"] = (
            f"Summed Histogram of time between visits at each point in the sky, " f"in {f} band(s)."
        )
        displayDict["order"] = filterorders[f]
        bundleList.append(
            mb.MetricBundle(
                m1,
                healpixslicer,
                constraint=filtersqls[f],
                info_label=filterinfo_label[f],
                run_name=runName,
                plot_dict=plotDict,
                plot_funcs=plotFuncs,
                display_dict=displayDict,
            )
        )

        gaps = [3.0, 7.0, 24.0]
        for gap in gaps:
            summary_stats = []
            summary_stats.append(
                metrics.AreaSummaryMetric(
                    area=18000,
                    reduce_func=np.median,
                    decreasing=True,
                    metric_name="Median N gaps in %s at %ihr in top 18k" % (f, gap),
                )
            )

            summary_stats.append(
                metrics.AreaSummaryMetric(
                    area=18000,
                    reduce_func=np.mean,
                    decreasing=True,
                    metric_name="Mean N gaps in %s at %ihr in top 18k" % (f, gap),
                )
            )

            summary_stats.append(metrics.MeanMetric())
            summary_stats.append(metrics.MedianMetric())

            m2 = metrics.GapsMetric(
                time_scale=gap,
                metric_name="Gaps_%ihr" % gap,
            )
            plotFuncs = [plots.HealpixSkyMap(), plots.HealpixHistogram()]
            plotDict = {"color_min": 0, "color": colors[f], "percentile_clip": 95}
            displayDict[
                "caption"
            ] = "Number of times the timescale of ~%i hours is sampled in %s band(s)." % (gap, f)
            displayDict["order"] = filterorders[f]
            bundleList.append(
                mb.MetricBundle(
                    m2,
                    healpixslicer,
                    constraint=filtersqls[f],
                    info_label=filterinfo_label[f],
                    run_name=runName,
                    summary_metrics=summary_stats,
                    plot_dict=plotDict,
                    plot_funcs=plotFuncs,
                    display_dict=displayDict,
                )
            )

    # FilterPairTgaps Metric (for TVS)
    m1 = maf.FilterPairTGapsMetric()
    plotDict = {"color_min": 0, "color_max": 1500, "x_min": 0, "x_max": 2000}
    displayDict["order"] = 0
    displayDict["caption"] = (
        "Evaluate the distribution of filter pairs and time gaps at each point in "
        "the sky. The time gaps are evaluated on a logarithmic spacing "
        "from 0-100 days for pairs of filters, and 0-3650 days for same filters."
    )
    summarystats = [
        metrics.MedianMetric(),
        metrics.MeanMetric(),
        metrics.PercentileMetric(percentile=80),
    ]
    bundle = maf.MetricBundle(
        m1,
        healpixslicer,
        None,
        run_name=runName,
        summary_metrics=summarystats,
        plot_dict=plotDict,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Presto KNe metric
    displayDict["group"] = "Variables/Transients"
    displayDict["subgroup"] = "Presto KNe"
    displayDict["caption"] = "Probability of detecting and classifying a KNe"
    prestoslicer = maf.generate_presto_pop_slicer(skyregion="extragalactic")
    metric = maf.PrestoColorKNePopMetric(
        skyregion="extragalactic",
        metric_name="PrestoKNe",
        mjd0=mjd0,
    )
    summaryMetrics_kne = [maf.MedianMetric(), maf.SumMetric()]
    bundleList.append(
        maf.MetricBundle(
            metric,
            prestoslicer,
            None,
            run_name=runName,
            display_dict=displayDict,
            summary_metrics=summaryMetrics_kne,
        )
    )

    # XRB metric
    displayDict["subgroup"] = "XRB"
    displayDict["order"] = 0
    displayDict["caption"] = "Number or characterization of XRBs."
    n_events = 10000
    xrbslicer = maf.generate_xrb_pop_slicer(n_events=n_events)
    metric = maf.XRBPopMetric(output_lc=False, mjd0=mjd0)
    xrb_summaryMetrics = [
        maf.SumMetric(metric_name="Total detected"),
        maf.CountMetric(metric_name="Total lightcurves in footprint"),
        maf.CountMetric(metric_name="Total lightcurves on sky", mask_val=0),
        maf.MeanMetric(metric_name="Fraction detected in footprint"),
        maf.MeanMetric(mask_val=0, metric_name="Fraction detected of total"),
        maf.MedianMetric(metric_name="Median"),
        maf.MeanMetric(metric_name="Mean"),
    ]

    bundleList.append(
        maf.MetricBundle(
            metric,
            xrbslicer,
            "",
            run_name=runName,
            summary_metrics=xrb_summaryMetrics,
            display_dict=displayDict,
        )
    )

    #########################
    #########################
    # Galactic Plane - TVS/MW
    #########################
    #########################

    displayDict = {"group": "Galactic Plane", "subgroup": ""}

    footprint_summaries = [metrics.SumMetric()]
    footprint_plotDicts = {"percentile_clip": 95}
    filter_summaries = [
        metrics.MeanMetric(),
        metrics.MedianMetric(),
        metrics.RmsMetric(),
        metrics.AreaThresholdMetric(lower_threshold=0.8),
    ]
    filter_plotdicts = {"color_min": 0, "color_max": 2, "x_min": 0, "x_max": 5}
    timescale_summaries = [
        metrics.SumMetric(),
        metrics.MedianMetric(),
        metrics.AreaThresholdMetric(lower_threshold=0.5),
    ]
    timescale_plotdicts = {"color_min": 0, "color_max": 1, "x_min": 0, "x_max": 1}

    galactic_plane_map_keys = maps.galplane_priority_map(nside=64, get_keys=True)
    science_maps = [
        s.replace("galplane_priority_", "").split(":")[0] for s in galactic_plane_map_keys if "sum" in s
    ]

    slicer = slicers.HealpixSlicer(nside=64, use_cache=False)
    sql = None
    bundles = {}
    for m in science_maps:
        displayDict["subgroup"] = m
        displayDict["caption"] = f"Footprint relevant for {m} galactic plane science case."
        footprintmetric = metrics.GalPlaneFootprintMetric(science_map=m)
        bundles[f"{m} footprint"] = mb.MetricBundle(
            footprintmetric,
            slicer,
            sql,
            plot_dict=footprint_plotDicts,
            run_name=runName,
            summary_metrics=footprint_summaries,
            display_dict=displayDict,
        )
        displayDict["caption"] = f"Filter balance in region of {m}  galactic plane science case."
        filtermetric = metrics.GalPlaneTimePerFilterMetric(science_map=m)
        bundles[f"{m} filter"] = mb.MetricBundle(
            filtermetric,
            slicer,
            sql,
            plot_dict=filter_plotdicts,
            run_name=runName,
            summary_metrics=filter_summaries,
            display_dict=displayDict,
        )
        displayDict["caption"] = f"Timescale intervals in region of {m} galactic plane science case."
        visit_timescalesmetric = metrics.GalPlaneVisitIntervalsTimescaleMetric(science_map=m)
        bundles[f"{m} visit intervals"] = mb.MetricBundle(
            visit_timescalesmetric,
            slicer,
            sql,
            plot_dict=timescale_plotdicts,
            run_name=runName,
            summary_metrics=timescale_summaries,
            display_dict=displayDict,
        )
        displayDict["caption"] = f"Seasonal coverage in region of {m} galactic plane science case."
        season_timescalemetric = metrics.GalPlaneSeasonGapsTimescaleMetric(science_map=m)
        bundles[f"{m} season gaps"] = mb.MetricBundle(
            season_timescalemetric,
            slicer,
            sql,
            plot_dict=timescale_plotdicts,
            run_name=runName,
            summary_metrics=timescale_summaries,
            display_dict=displayDict,
        )

    bundleList += list(bundles.values())

    #########################
    #########################
    # Milky Way
    #########################
    #########################

    displayDict = {"group": "Milky Way", "subgroup": ""}

    displayDict["subgroup"] = "N stars"
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    sum_stats = [metrics.SumMetric(metric_name="Total N Stars, crowding")]
    for f in filterlist:
        stellar_map = maps.StellarDensityMap(filtername=f)
        displayDict["order"] = filterorders[f]
        displayDict["caption"] = (
            "Number of stars in %s band with an measurement uncertainty due to crowding "
            "of less than 0.2 mag" % f
        )
        # Configure the NstarsMetric - note 'filtername' refers to the
        # filter in which to evaluate crowding
        metric = metrics.NstarsMetric(
            crowding_error=0.2,
            filtername=f,
            ignore_crowding=False,
            maps=[],
        )
        plotDict = {"n_ticks": 5, "log_scale": True, "color_min": 100}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            filtersqls[f],
            run_name=runName,
            summary_metrics=sum_stats,
            plot_funcs=subsetPlots,
            plot_dict=plotDict,
            display_dict=displayDict,
            maps_list=[stellar_map],
        )
        bundleList.append(bundle)

    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    sum_stats = [metrics.SumMetric(metric_name="Total N Stars, no crowding")]
    for f in filterlist:
        stellar_map = maps.StellarDensityMap(filtername=f)
        displayDict["order"] = filterorders[f]
        displayDict["caption"] = (
            "Number of stars in %s band with an measurement uncertainty "
            "of less than 0.2 mag, not considering crowding" % f
        )
        # Configure the NstarsMetric - note 'filtername' refers to the
        # filter in which to evaluate crowding
        metric = metrics.NstarsMetric(
            crowding_error=0.2,
            filtername=f,
            ignore_crowding=True,
            metric_name="Nstars_no_crowding",
            maps=[],
        )
        plotDict = {"n_ticks": 5, "log_scale": True, "color_min": 100}
        bundle = mb.MetricBundle(
            metric,
            slicer,
            filtersqls[f],
            run_name=runName,
            summary_metrics=sum_stats,
            plot_funcs=subsetPlots,
            plot_dict=plotDict,
            display_dict=displayDict,
            maps_list=[stellar_map],
        )
        bundleList.append(bundle)

    # Brown Dwarf Volume
    displayDict["subgroup"] = "Brown Dwarf"
    displayDict["order"] = 0
    l7_bd_mags = {"i": 20.09, "z": 18.18, "y": 17.13}
    displayDict["caption"] = (
        f"The expected maximum distance at which an L7 brown dwarf with magnitude {l7_bd_mags} "
        f"would have a parallax SNR of 10.0. The summary statistic represents the volume enclosed by "
        f"the result of this metric (BDParallaxMetric)."
    )
    sum_stats = [metrics.VolumeSumMetric(nside=nside)]
    metric = metrics.BDParallaxMetric(mags=l7_bd_mags, metric_name="Brown Dwarf, L7")
    sql = ""
    plotDict = {}
    bundleList.append(
        mb.MetricBundle(
            metric,
            healpixslicer,
            sql,
            plot_dict=plotDict,
            summary_metrics=sum_stats,
            display_dict=displayDict,
            run_name=runName,
        )
    )

    l4_bd_mags = {"i": 18.35, "z": 16.68, "y": 15.66}
    displayDict["caption"] = (
        f"The expected maximum distance at which an L4 brown dwarf with magnitude {l4_bd_mags} "
        f"would have a parallax SNR of 10.0. The summary statistic represents the total volume enclosed "
        f"by the result of this metric (BDParallaxMetric)."
    )
    metric = metrics.BDParallaxMetric(mags=l4_bd_mags, metric_name="Brown Dwarf, L4")
    bundleList.append(
        mb.MetricBundle(
            metric,
            healpixslicer,
            sql,
            plot_dict=plotDict,
            summary_metrics=sum_stats,
            display_dict=displayDict,
            run_name=runName,
        )
    )

    displayDict["subgroup"] = "Young Stellar Objects"
    displayDict["caption"] = (
        "The number of expected Young Stellar Objects with age t<10 Myr and "
        "mass >0.3 solar masses, using coadded depths in g, r and i bands."
    )
    # The underlying dustmap is nside=64, but can be resampled
    nside_yso = 64
    dustmap3d = maps.DustMap3D(nside=nside_yso)
    sql = ""
    # Let's plug in the magnitudes for one type
    metric = maf.maf_contrib.NYoungStarsMetric()
    slicer = maf.slicers.HealpixSlicer(nside=nside_yso, use_cache=False)
    summaryStats = [maf.metrics.SumMetric()]
    plotDict = {"log_scale": True, "color_min": 1}
    bundleList.append(
        maf.metric_bundles.MetricBundle(
            metric,
            slicer,
            sql,
            maps_list=[dustmap3d],
            plot_dict=plotDict,
            summary_metrics=summaryStats,
            run_name=runName,
            display_dict=displayDict,
        )
    )

    ## Local Volume Dwarf Satellites
    displayDict["group"] = "Local Volume"
    displayDict["subgroup"] = "LV dwarf satellites"
    displayDict["order"] = 0
    # First the known dwarf satellite galaxies
    displayDict["caption"] = (
        "Predicted magnitude detection limit (including stellar density and "
        "star-galaxy separation) at the location of known LV dwarf galaxies. "
    )
    i_starMap = maf.maps.StellarDensityMap(filtername="i")
    lv_slicer = maf.maf_contrib.generate_known_lv_dwarf_slicer()
    lv_metric = maf.maf_contrib.LVDwarfsMetric()
    sqlconstraint = '(filter = "i" OR filter = "g")'
    info_label = "gi"
    cutoff = -6.4
    summary_metrics = [
        maf.metrics.CountBeyondThreshold(lower_threshold=cutoff, metric_name=f"Total detected {cutoff}"),
        maf.metrics.CountBeyondThreshold(lower_threshold=-7.0, metric_name="Total detected -7.0"),
    ]
    plotDict = {"n_ticks": 7}
    bundle = maf.MetricBundle(
        lv_metric,
        lv_slicer,
        sqlconstraint,
        maps_list=[i_starMap],
        summary_metrics=summary_metrics,
        info_label=info_label,
        display_dict=displayDict,
        plot_dict=plotDict,
    )
    bundleList.append(bundle)

    displayDict["order"] += 1
    displayDict["caption"] = (
        "Predicted magnitude detection limit (including stellar density and "
        "star-galaxy separation), over the whole sky to a distance of 4 Mpc."
    )
    lv_metric2 = maf.maf_contrib.LVDwarfsMetric(distlim=4.0 * u.Mpc)  # for a distance limit, healpix map
    dustmap = maf.maps.DustMap(nside=32)
    lv_healpix_slicer = maf.slicers.HealpixSlicer(nside=32, use_cache=False)
    summary_area = maf.AreaThresholdMetric(lower_threshold=cutoff, metric_name=f"Area M_v>{cutoff}")
    bundle = maf.MetricBundle(
        lv_metric2,
        lv_healpix_slicer,
        sqlconstraint,
        maps_list=[i_starMap, dustmap],
        info_label=info_label,
        display_dict=displayDict,
        summary_metrics=summary_area,
        plot_dict=plotDict,
    )
    bundleList.append(bundle)

    displayDict["order"] += 1
    displayDict["caption"] = (
        "Predicted magnitude detection limit (including stellar density and "
        "star-galaxy separation), over the southern celestial pole,"
        "to a distance of 0.1 Mpc."
    )
    sqlconstraint = '(filter = "i" OR filter = "g") and fieldDec < -60'
    info_label = "gi SCP"
    lv_metric3 = maf.maf_contrib.LVDwarfsMetric(distlim=0.1 * u.Mpc)  # for a distance limit, healpix map
    summary_area = maf.metrics.AreaThresholdMetric(lower_threshold=0.0, metric_name="Area M_v>0.0")
    summary_median = maf.metrics.MedianMetric()
    bundle = maf.MetricBundle(
        lv_metric3,
        lv_healpix_slicer,
        sqlconstraint,
        maps_list=[i_starMap, dustmap],
        info_label=info_label,
        display_dict=displayDict,
        plot_dict=plotDict,
        summary_metrics=[summary_area, summary_median],
    )
    bundleList.append(bundle)

    #########################
    #########################
    # Per year things
    #########################
    #########################

    plotDict = {}
    night_cuttoffs = np.arange(1, 11, 1) * 365.25
    slicer = slicers.HealpixSlicer(nside=nside)
    for i, cuttoff in enumerate(night_cuttoffs):
        for filtername in "ugrizy":
            sql = "night<%i and filter='%s'" % (cuttoff, filtername)
            metric = metrics.Coaddm5Metric(metric_name="coadd %s, year<%i" % (filtername, i + 1))
            displayDict = {"group": "Per year", "subgroup": "Coadds"}
            bundle = mb.MetricBundle(
                metric,
                slicer,
                sql,
                run_name=runName,
                summary_metrics=standardStats,
                plot_funcs=subsetPlots,
                plot_dict=plotDict,
                display_dict=displayDict,
            )
            bundleList.append(bundle)

            metric = metrics.SnrWeightedMetric(
                col="seeingFwhmEff",
                metric_name="SNR-weighted FWHMeff %s, year<%i" % (filtername, i + 1),
            )
            displayDict = {"group": "Per year", "subgroup": "Seeing"}
            bundle = mb.MetricBundle(
                metric,
                slicer,
                sql,
                run_name=runName,
                summary_metrics=standardStats,
                plot_funcs=subsetPlots,
                plot_dict=plotDict,
                display_dict=displayDict,
            )
            bundleList.append(bundle)

    #########################
    #########################
    # Scaling numbers
    #########################
    #########################

    displayDict = {"group": "Scaling Numbers", "subgroup": ""}
    displayDict["subgroup"] = "N gals"
    sql = 'filter="i"'
    metric = metrics.NgalScaleMetric()
    # galaxy counting uses dustmap
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    displayDict["caption"] = (
        "Approximate number of resolvable galaxies in i band, scaled by the "
        "coadded depth and median seeing. A dust and magnitude cut has been applied."
    )
    bundle = mb.MetricBundle(
        metric,
        slicer,
        sql,
        run_name=runName,
        summary_metrics=[metrics.SumMetric()],
        plot_funcs=subsetPlots,
        plot_dict=plotDict,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    displayDict["subgroup"] = "Lightcurve Pts"
    metric = metrics.NlcPointsMetric(nside=nside)
    # NlcPoints metric uses star density maps
    slicer = slicers.HealpixSlicer(nside=nside, use_cache=False)
    displayDict["caption"] = (
        "Approximate number of expected stellar measurements (nstars * nobs) "
        "in all filters, where the limiting magnitude for at least 10 visits "
        "is fainter than 21st magnitude, using a TRILEGAL stellar density map."
    )
    bundle = mb.MetricBundle(
        metric,
        slicer,
        None,
        run_name=runName,
        summary_metrics=[metrics.SumMetric()],
        plot_funcs=subsetPlots,
        plot_dict=plotDict,
        display_dict=displayDict,
    )
    bundleList.append(bundle)

    # Set the run_name for all bundles and return the bundleDict.
    for b in bundleList:
        b.set_run_name(runName)
    bundleDict = mb.make_bundles_dict_from_list(bundleList)

    return bundleDict



__all__ = ("UseMetric",)

import numpy as np

from rubin_sim.maf.metrics.base_metric import BaseMetric


class UseMetric(BaseMetric):  # pylint: disable=too-few-public-methods
    """Metric to classify visits by type of visits"""

    def __init__(self, note_col="note", **kwargs):
        self.note_col = note_col
        super().__init__(col=[note_col], metric_dtype="object", **kwargs)

    def run(self, data_slice, slice_point=None):  # pylint: disable=invalid-name
        """Run the metric.

        Parameters
        ----------
        data_slice : `np.ndarray`, (N,)`
        slice_point : `dict`
           Dictionary of slice_point metadata passed to each metric.
           E.g. the ra/dec of the healpix pixel.

        Returns
        -------
        use_name : `str`
            use at each slice_point.
        """
        use_name = None
        visible_bands = ("u", "g", "r")
        notes = data_slice[self.note_col]
        if len(notes.shape) == 0:
            note = notes
        else:
            note = notes[0]
            assert np.all(notes == note)

        note_elems = note.replace(":", ", ").split(", ")
        # XXX--survey note strings should not be hard-coded here.
        if note_elems[0] == "greedy":
            use_name = note_elems[0]
        if note_elems[0] == "DD":
            use_name = note_elems[1]
        if (note_elems[0] == "blob") | (note_elems[0] == "blob_twi"):
            use_name = "wide with only IR"
            for band in visible_bands:
                if band in note_elems[1]:
                    use_name = "wide with u, g, or r"

        assert use_name is not None, f"Unrecognized note: {note}"
        return use_name


# internal functions & classes



# Example for numObsInSurveyTimeOverlap
# Somayeh Khakpash - Lehigh University
# Last edited : 10/21/2020
# Calculates number of observations during simultaneous windows of another survey.
# SurveyObsWin is the list of the survey observing window/inter-seasonal gap intervals. It should be in the format:
# SurveyObsWin = [ [YYYY-MM-DD, YYYY-MM-DD] , [YYYY-MM-DD, YYYY-MM-DD] , ... , [YYYY-MM-DD, YYYY-MM-DD] ]

__all__ = ("NumObsInSurveyTimeOverlapMetric",)

import numpy as np
from astropy.time import Time

from rubin_sim.maf.metrics import BaseMetric


class NumObsInSurveyTimeOverlapMetric(BaseMetric):
    def __init__(
        self,
        survey_obs_win,
        time_col="observationStartMJD",
        metric_name="NumObsInSurveyTimeOverlapMetric",
        **kwargs,
    ):
        self.time_col = time_col
        self.metric_name = metric_name
        self.survey_obs_win = survey_obs_win
        super(NumObsInSurveyTimeOverlapMetric, self).__init__(col=time_col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        n__obs = 0
        for interval in self.survey_obs_win:
            start_interval = Time(interval[0] + " 00:00:00")
            end_interval = Time(interval[1] + " 00:00:00")
            index = np.where(
                (data_slice[self.time_col] > start_interval.mjd)
                & (data_slice[self.time_col] < end_interval.mjd)
            )[0]
            n__obs = n__obs + np.size(index)

        return n__obs



import shutil
import unittest

from rubin_sim.maf.db import ResultsDb
from rubin_sim.maf.run_comparison import combine_result_dbs


class TestGather(unittest.TestCase):
    def setUp(self):
        # Make two results database to access
        results_db = ResultsDb(out_dir="temp1")
        results_db.update_metric(
            "test_metric_name",
            "test_slicer_name",
            "temp",
            "test_sql_constraint",
            "test_metric_info_label",
            "test_metric_datafile",
        )

        results_db.update_summary_stat(1, "test_summary_name", summary_value=100.0)

        results_db = ResultsDb(out_dir="temp2")
        results_db.update_metric(
            "test_metric_name",
            "test_slicer_name",
            "temp",
            "test_sql_constraint",
            "test_metric_info_label",
            "test_metric_datafile",
        )

        results_db.update_summary_stat(1, "test_summary_name", summary_value=200.0)

    def tearDown(self):
        shutil.rmtree("temp1")
        shutil.rmtree("temp2")

    def test_gather_summaries(self):
        # Gather the results into a single db
        df = combine_result_dbs(run_dirs=["temp1", "temp2"])
        # Check that values are correct
        assert df.loc["temp1"].max() == 100.0
        assert df.loc["temp2"].max() == 200.0


if __name__ == "__main__":
    unittest.main()



__all__ = ("Orbits",)

import warnings

import numpy as np
import pandas as pd


class Orbits:
    """Orbits reads, checks for required values, and stores orbit parameters for moving objects.

    Instantiate the class and then use read_orbits or set_orbits to set the orbit values.

    self.orbits stores the orbital parameters, as a pandas dataframe.
    self.dataCols defines the columns required, although obj_id, H, g, and sed_filename are optional.
    """

    def __init__(self):
        self.orbits = None
        self.orb_format = None

        # Specify the required columns/values in the self.orbits dataframe.
        # Which columns are required depends on self.orb_format.
        self.data_cols = {}
        self.data_cols["COM"] = [
            "obj_id",
            "q",
            "e",
            "inc",
            "Omega",
            "argPeri",
            "tPeri",
            "epoch",
            "H",
            "g",
            "sed_filename",
        ]
        self.data_cols["KEP"] = [
            "obj_id",
            "a",
            "e",
            "inc",
            "Omega",
            "argPeri",
            "meanAnomaly",
            "epoch",
            "H",
            "g",
            "sed_filename",
        ]
        self.data_cols["CAR"] = [
            "obj_id",
            "x",
            "y",
            "z",
            "xdot",
            "ydot",
            "zdot",
            "epoch",
            "H",
            "g",
            "sed_filename",
        ]

    def __len__(self):
        return len(self.orbits)

    def __getitem__(self, i):
        orb = Orbits()
        orb.set_orbits(self.orbits.iloc[i])
        return orb

    def __iter__(self):
        for i, orbit in self.orbits.iterrows():
            orb = Orbits()
            orb.set_orbits(orbit)
            yield orb

    def __eq__(self, other_orbits):
        if isinstance(other_orbits, Orbits):
            if self.orb_format != other_orbits.orb_format:
                return False
            for col in self.data_cols[self.orb_format]:
                if not np.all(self.orbits[col].values == other_orbits.orbits[col].values):
                    return False
                else:
                    return True
        else:
            return False

    def __neq__(self, other_orbits):
        if self == other_orbits:
            return False
        else:
            return True

    def set_orbits(self, orbits):
        """Set and validate orbital parameters contain all required values.

        Sets self.orbits and self.orb_format.
        If objid is not present in orbits, a sequential series of integers will be used.
        If H is not present in orbits, a default value of 20 will be used.
        If g is not present in orbits, a default value of 0.15 will be used.
        If sed_filename is not present in orbits, either C or S type will be assigned,
        according to the semi-major axis value.

        Parameters
        ----------
        orbits : `pd.DataFrame`, `pd.Series` or `np.ndarray`
           Array-like object containing orbital parameter information.
        """
        # Do we have a single item or multiples?
        if isinstance(orbits, pd.Series):
            # Passed a single SSO in Series, convert to a DataFrame.
            orbits = pd.DataFrame([orbits])
        elif isinstance(orbits, np.ndarray):
            # Passed a numpy array, convert to DataFrame.
            orbits = pd.DataFrame.from_records(orbits)
        elif isinstance(orbits, np.record):
            # This was a single object in a numpy array and we should be a bit fancy.
            orbits = pd.DataFrame.from_records([orbits], columns=orbits.dtype.names)
        elif isinstance(orbits, pd.DataFrame):
            # This was a pandas dataframe .. but we probably want to drop the index and recount.
            orbits.reset_index(drop=True, inplace=True)

        if "index" in orbits:
            del orbits["index"]

        n_sso = len(orbits)

        # Error if orbits is empty (this avoids hard-to-interpret error messages from pyoorb).
        if n_sso == 0:
            raise ValueError("Length of the orbits dataframe was 0.")

        # Discover which type of orbital parameters we have on disk.
        self.orb_format = None
        if "FORMAT" in orbits:
            if ~(orbits["FORMAT"] == orbits["FORMAT"].iloc[0]).all():
                raise ValueError("All orbital elements in the set should have the same FORMAT.")
            self.orb_format = orbits["FORMAT"].iloc[0]
            # Backwards compatibility .. a bit. CART is deprecated, so swap it to CAR.
            if self.orb_format == "CART":
                self.orb_format = "CAR"
            del orbits["FORMAT"]
            # Check that the orbit format is approximately right.
            if self.orb_format == "COM":
                if "q" not in orbits:
                    raise ValueError('The stated format was COM, but "q" not present in orbital elements?')
            if self.orb_format == "KEP":
                if "a" not in orbits:
                    raise ValueError('The stated format was KEP, but "a" not present in orbital elements?')
            if self.orb_format == "CAR":
                if "x" not in orbits:
                    raise ValueError('The stated format was CAR but "x" not present in orbital elements?')
        if self.orb_format is None:
            # Try to figure out the format, if it wasn't provided.
            if "q" in orbits:
                self.orb_format = "COM"
            elif "a" in orbits:
                self.orb_format = "KEP"
            elif "x" in orbits:
                self.orb_format = "CAR"
            else:
                raise ValueError(
                    "Can't determine orbital type, as neither q, a or x in input orbital elements.\n"
                    "Was attempting to base orbital element quantities on header row, "
                    "with columns: \n%s" % orbits.columns
                )

        # Check that the orbit epoch is within a 'reasonable' range, to detect possible column mismatches.
        general_epoch = orbits["epoch"].head(1).values[0]
        # Look for epochs between 1800 and 2200 - this is primarily to check if people used MJD (and not JD).
        expect_min_epoch = -21503.0
        expect_max_epoch = 124594.0
        if general_epoch < expect_min_epoch or general_epoch > expect_max_epoch:
            raise ValueError(
                "The epoch detected for this orbit is odd - %f. "
                "Expecting a value between %.1f and %.1f (MJD!)"
                % (general_epoch, expect_min_epoch, expect_max_epoch)
            )

        # If these columns are not available in the input data, auto-generate them.
        if "obj_id" not in orbits:
            obj_id = np.arange(0, n_sso, 1)
            orbits = orbits.assign(obj_id=obj_id)
        if "H" not in orbits:
            orbits = orbits.assign(H=20.0)
        if "g" not in orbits:
            orbits = orbits.assign(g=0.15)
        if "sed_filename" not in orbits:
            orbits = orbits.assign(sed_filename=self.assign_sed(orbits))

        # Make sure we gave all the columns we need.
        for col in self.data_cols[self.orb_format]:
            if col not in orbits:
                raise ValueError(
                    "Missing required orbital element %s for orbital format type %s" % (col, self.orb_format)
                )

        # Check to see if we have duplicates.
        if len(orbits["obj_id"].unique()) != n_sso:
            warnings.warn(
                "There are duplicates in the orbit obj_id values" + " - was this intended? (continuing)."
            )
        # All is good.
        self.orbits = orbits

    def assign_sed(self, orbits, random_seed=None):
        """Assign either a C or S type SED, depending on the semi-major axis of the object.
        P(C type) = 0 (a<2); 0.5*a - 1 (2<a<4); 1 (a > 4),
        based on figure 23 from Ivezic et al 2001 (AJ, 122, 2749).

        Parameters
        ----------
        orbits : `pd.DataFrame`, `pd.Series` or `np.ndarray`
           Array-like object containing orbital parameter information.

        Returns
        -------
        sedvals : `np.ndarray`
            Array containing the SED type for each object in 'orbits'.
        """
        # using fig. 23 from Ivezic et al. 2001 (AJ, 122, 2749),
        # we can approximate the sed types with a simple linear form:
        #  p(C) = 0 for a<2
        #  p(C) = 0.5*a-1  for 2<a<4
        #  p(C) = 1 for a>4
        # where a is semi-major axis, and p(C) is the probability that
        # an asteroid is C type, with p(S)=1-p(C) for S types.
        if "a" in orbits:
            a = orbits["a"]
        elif "q" in orbits:
            a = orbits["q"] / (1 - orbits["e"])
        elif "x" in orbits:
            # This definitely isn't right, but it's a placeholder to make it work for now.
            a = np.sqrt(orbits["x"] ** 2 + orbits["y"] ** 2 + orbits["z"] ** 2)
        else:
            raise ValueError("Need either a or q (plus e) in orbit data frame.")

        if not hasattr(self, "_rng"):
            if random_seed is not None:
                self._rng = np.random.RandomState(random_seed)
            else:
                self._rng = np.random.RandomState(42)

        chance = self._rng.random_sample(len(orbits))
        prob_c = 0.5 * a - 1.0
        # if chance <= prob_c:
        sedvals = np.where(chance <= prob_c, "C.dat", "S.dat")
        return sedvals

    def read_orbits(self, orbit_file, delim=None, skiprows=None):
        """Read orbits from a file, generating a pandas dataframe containing columns matching dataCols,
        for the appropriate orbital parameter format (currently accepts COM, KEP or CAR formats).

        After reading and standardizing the column names, calls self.set_orbits to validate the
        orbital parameters. Expects angles in orbital element formats to be in degrees.

        Note that readOrbits uses pandas.read_csv to read the data file with the orbital parameters.
        Thus, it should have column headers specifying the column names ..
        unless skiprows = -1 or there is just no header line at all.
        in which case it is assumed to be a standard DES format file, with no header line.

        Parameters
        ----------
        orbit_file : `str`
            The name of the input file containing orbital parameter information.
        delim : `str`, optional
            The delimiter for the input orbit file. Default is None, will use delim_whitespace=True.
        skiprows : `int`, optional
            The number of rows to skip before reading the header information for pandas.
            Default is None, which will trigger a check of the file to look for the header columns.
        """
        names = None

        # If skiprows is set, then we will assume the user has handled this so that the
        # first line read has the header information.
        # But, if skiprows is not set, then we have to do some checking to see if there is
        # header information and which row it might start in.
        if skiprows is None:
            skiprows = -1
            # Figure out whether the header is in the first line, or if there are rows to skip.
            # We need to do a bit of juggling to do this before pandas reads the whole orbit file though.
            with open(orbit_file, "r") as fp:
                headervalues = None
                for line in fp:
                    values = line.split()
                    try:
                        # If it is a valid orbit line, we expect column 3 to be a number.
                        float(values[3])
                        # And if it worked, we're done here (it's an orbit) - go on to parsing header values.
                        break
                    except (ValueError, IndexError):
                        # This wasn't a valid number or there wasn't anything in the third value.
                        # So this is either the header line or it's a comment line before the header columns.
                        skiprows += 1
                        headervalues = values

            if headervalues is not None:  # (and skiprows > -1)
                # There is a header, but we also need to check if there is a comment key at the start
                # of the proper header line.
                # ... Because this varies as well, and is sometimes separated from header columns.
                linestart = headervalues[0]
                if linestart == "#" or linestart == "!!" or linestart == "##":
                    names = headervalues[1:]
                else:
                    names = headervalues
                # Add 1 to skiprows, so that we skip the header column line.
                skiprows += 1

        # So now skiprows is a value. If it is -1, then there is no header information.
        if skiprows == -1:
            # No header; assume it's a typical DES file -
            # we'll assign the column names based on the FORMAT.
            names_com = (
                "obj_id",
                "FORMAT",
                "q",
                "e",
                "i",
                "node",
                "argperi",
                "t_p",
                "H",
                "epoch",
                "INDEX",
                "N_PAR",
                "MOID",
                "COMPCODE",
            )
            names_kep = (
                "obj_id",
                "FORMAT",
                "a",
                "e",
                "i",
                "node",
                "argperi",
                "meanAnomaly",
                "H",
                "epoch",
                "INDEX",
                "N_PAR",
                "MOID",
                "COMPCODE",
            )
            names_car = (
                "obj_id",
                "FORMAT",
                "x",
                "y",
                "z",
                "xdot",
                "ydot",
                "zdot",
                "H",
                "epoch",
                "INDEX",
                "N_PAR",
                "MOID",
                "COMPCODE",
            )
            # First use names_com, and then change if required.
            orbits = pd.read_csv(orbit_file, delim_whitespace=True, header=None, names=names_com)

            if orbits["FORMAT"][0] == "KEP":
                orbits.columns = names_kep
            elif orbits["FORMAT"][0] == "CAR":
                orbits.columns = names_car

        else:
            if delim is None:
                orbits = pd.read_csv(orbit_file, delim_whitespace=True, skiprows=skiprows, names=names)
            else:
                orbits = pd.read_csv(orbit_file, sep=delim, skiprows=skiprows, names=names)

        # Drop some columns that are typically present in DES files but that we don't need.
        if "INDEX" in orbits:
            del orbits["INDEX"]
        if "N_PAR" in orbits:
            del orbits["N_PAR"]
        if "MOID" in orbits:
            del orbits["MOID"]
        if "COMPCODE" in orbits:
            del orbits["COMPCODE"]
        if "tmp" in orbits:
            del orbits["tmp"]

        # Normalize the column names to standard values and identify the orbital element types.
        sso_cols = orbits.columns.values.tolist()

        # These are the alternative possibilities for various column headers
        # (depending on file version, origin, etc.)
        # that might need remapping from the on-file values to our standardized values.
        alt_names = {}
        alt_names["obj_id"] = [
            "obj_id",
            "objid",
            "!!ObjID",
            "!!OID",
            "!!S3MID",
            "OID",
            "S3MID" "objid(int)",
            "full_name",
            "#name",
        ]
        alt_names["q"] = ["q"]
        alt_names["a"] = ["a"]
        alt_names["e"] = ["e", "ecc"]
        alt_names["inc"] = ["inc", "i", "i(deg)", "incl"]
        alt_names["Omega"] = [
            "Omega",
            "omega",
            "node",
            "om",
            "node(deg)",
            "BigOmega",
            "Omega/node",
            "longNode",
        ]
        alt_names["argPeri"] = [
            "argPeri",
            "argperi",
            "omega/argperi",
            "w",
            "argperi(deg)",
            "peri",
        ]
        alt_names["tPeri"] = ["tPeri", "t_p", "timeperi", "t_peri", "T_peri"]
        alt_names["epoch"] = ["epoch", "t_0", "Epoch", "epoch_mjd"]
        alt_names["H"] = ["H", "magH", "magHv", "Hv", "H_v"]
        alt_names["g"] = ["g", "phaseV", "phase", "gV", "phase_g", "G"]
        alt_names["meanAnomaly"] = ["meanAnomaly", "meanAnom", "M", "ma"]
        alt_names["sed_filename"] = ["sed_filename", "sed"]
        alt_names["xdot"] = ["xdot", "xDot"]
        alt_names["ydot"] = ["ydot", "yDot"]
        alt_names["zdot"] = ["zdot", "zDot"]

        # Update column names that match any of the alternatives above.
        for name, alternatives in alt_names.items():
            intersection = list(set(alternatives) & set(sso_cols))
            if len(intersection) > 1:
                raise ValueError(
                    "Received too many possible matches to %s in orbit file %s" % (name, orbit_file)
                )
            if len(intersection) == 1:
                idx = sso_cols.index(intersection[0])
                sso_cols[idx] = name
        # Assign the new column names back to the orbits dataframe.
        orbits.columns = sso_cols

        # Failing on negaitive inclinations.
        if "inc" in orbits.keys():
            if np.min(orbits["inc"]) < 0:
                negative_incs = np.where(orbits["inc"].values < 0)[0]
                negative_ids = orbits["obj_id"].values[negative_incs]
                ValueError("Negative orbital inclinations not supported. Problem obj_ids=%s" % negative_ids)

        # Validate and assign orbits to self.
        self.set_orbits(orbits)

    def update_orbits(self, neworb):
        """Update existing orbits with new values, leaving OrbitIds, H, g, and sed_filenames in place.

        Example use: transform orbital parameters (using PyOrbEphemerides) and then replace original values.
        Example use: propagate orbital parameters (using PyOrbEphemerides) and then replace original values.

        Parameters
        ----------
        neworb: `pd.DataFrame`
        """
        col_orig = ["obj_id", "sed_filename"]
        new_order = ["obj_id"] + [n for n in neworb.columns] + ["sed_filename"]
        updated_orbits = neworb.join(self.orbits[col_orig])[new_order]
        self.set_orbits(updated_orbits)



# UniSlicer class.
# This slicer simply returns the indexes of all data points. No slicing done at all.
__all__ = ("UniSlicer",)

from functools import wraps

import numpy as np

from .base_slicer import BaseSlicer


class UniSlicer(BaseSlicer):
    """UniSlicer."""

    def __init__(self, verbose=True, badval=-666):
        """Instantiate unislicer."""
        super(UniSlicer, self).__init__(verbose=verbose, badval=badval)
        self.nslice = 1
        self.shape = self.nslice
        self.slice_points["sid"] = np.array(
            [
                0,
            ],
            int,
        )
        self.plot_funcs = []

    def setup_slicer(self, sim_data, maps=None):
        """Use sim_data to set indexes to return."""
        self._run_maps(maps)
        sim_dataCol = sim_data.dtype.names[0]
        self.indices = np.ones(len(sim_data[sim_dataCol]), dtype="bool")

        @wraps(self._slice_sim_data)
        def _slice_sim_data(islice):
            """Return all indexes in sim_data."""
            idxs = self.indices
            return {"idxs": idxs, "slice_point": {"sid": islice}}

        setattr(self, "_slice_sim_data", _slice_sim_data)

    def __eq__(self, other_slicer):
        """Evaluate if slicers are equivalent."""
        if isinstance(other_slicer, UniSlicer):
            return True
        else:
            return False



__all__ = ("GalCoordsMap",)

from astropy import units as u
from astropy.coordinates import SkyCoord

from rubin_sim.maf.maps import BaseMap


class GalCoordsMap(BaseMap):
    def __init__(self):
        self.keynames = ["gall", "galb"]

    def run(self, slice_points):
        coords = SkyCoord(ra=slice_points["ra"] * u.rad, dec=slice_points["dec"] * u.rad)
        gal = coords.galactic
        gall = gal.l.rad
        galb = gal.b.rad
        slice_points["gall"] = gall
        slice_points["galb"] = galb
        return slice_points



__all__ = (
    "fwhm_eff2_fwhm_geom",
    "fwhm_geom2_fwhm_eff",
    "calc_neff",
    "calc_instr_noise_sq",
    "calc_total_non_source_noise_sq",
    "calc_snr_sed",
    "calc_m5",
    "calc_sky_counts_per_pixel_for_m5",
    "calc_gamma",
    "calc_snr_m5",
    "calc_astrometric_error",
    "mag_error_from_snr",
    "calc_mag_error_m5",
    "calc_mag_error_sed",
)

import numpy

from .sed import Sed


def fwhm_eff2_fwhm_geom(fwhm_eff):
    """Convert fwhm_eff to fwhm_geom.

    This conversion was calculated by Bo Xin and Zeljko Ivezic
    (and will be in an update on the LSE-40 and overview papers).

    Parameters
    ----------
    fwhm_eff: `float`
        the single-gaussian equivalent FWHM value, appropriate for calc_neff,
        in arcseconds

    Returns
    -------
    fwhm_geom : `float`
        FWHM geom, the geometric FWHM value as measured from a typical
        PSF profile in arcseconds.
    """
    fwhm_geom = 0.822 * fwhm_eff + 0.052
    return fwhm_geom


def fwhm_geom2_fwhm_eff(fwhm_geom):
    """Convert fwhm_geom to fwhm_eff.

    This conversion was calculated by Bo Xin and Zeljko Ivezic
    (and will be in an update on the LSE-40 and overview papers).

    Parameters
    ----------
    fwhm_geom: `float`
        The geometric FWHM value, as measured from a typical PSF profile,
         in arcseconds.

    Returns
    -------
    fwhm_eff: `float`
        FWHM effective, the single-gaussian equivalent FWHM value,
        appropriate for calc_neff, in arcseconds.
    """
    fwhm_eff = (fwhm_geom - 0.052) / 0.822
    return fwhm_eff


def calc_neff(fwhm_eff, platescale):
    """Calculate the effective number of pixels in a single gaussian PSF.

    This equation comes from LSE-40, equation 27.
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    Parameters
    ----------
    fwhm_eff: `float`
        The width of a single-gaussian that produces correct
        Neff for typical PSF profile.
    platescale: `float`
        The platescale in arcseconds per pixel (0.2 for LSST)

    Returns
    -------
    nEff : `float`
        The effective number of pixels contained in the PSF

    Notes
    -----
    The fwhm_eff is a way to represent the equivalent seeing value, if the
    atmosphere could be simply represented as a single gaussian (instead of
    a more complicated von Karman profile for the atmosphere, convolved
    properly with the telescope hardware additional blurring of 0.4").
    A translation from the geometric FWHM to the fwhm_eff is provided
    in fwhm_geom2_fwhm_eff.
    """
    return 2.266 * (fwhm_eff / platescale) ** 2


def calc_instr_noise_sq(phot_params):
    """Combine all of the noise due to intrumentation into one value

    Parameters
    ----------
    phot_params : `PhotometricParameters`
        A PhotometricParameters object that carries details about the
        photometric response of the telescope.

    Returns
    -------
    inst_noise_sq : `float`
        The noise due to all of these sources added in quadrature in ADU counts
    """

    # instrumental squared noise in electrons
    inst_noise_sq = (
        phot_params.nexp * phot_params.readnoise**2
        + phot_params.darkcurrent * phot_params.exptime * phot_params.nexp
        + phot_params.nexp * phot_params.othernoise**2
    )

    # convert to ADU counts
    inst_noise_sq = inst_noise_sq / (phot_params.gain * phot_params.gain)

    return inst_noise_sq


def calc_total_non_source_noise_sq(sky_sed, hardwarebandpass, phot_params, fwhm_eff):
    """Calculate the noise due to instrumentation and sky background.

    Parameters
    ----------
    sky_sed : `Sed`
        A Sed object representing the sky (normalized so that
        sky_sed.calc_mag() gives the sky brightness in
        magnitudes per square arcsecond)
    hardwarebandpass : `Bandpass`
        A Bandpass object containing just the instrumentation
        throughputs (no atmosphere)
    phot_params : `PhotometricParameters`
        A PhotometricParameters object containing information
        about the photometric properties of the telescope.
    fwhm_eff : `float`
        fwhm_eff in arcseconds

    Returns
    -------
    total_noise_sq : `float`
        total non-source noise squared (in ADU counts)
        (this is simga^2_tot * neff in equation 41 of the SNR document
        https://ls.st/LSE-40 )
    """

    # Calculate the effective number of pixels for double-Gaussian PSF
    neff = calc_neff(fwhm_eff, phot_params.platescale)

    # Calculate the counts from the sky.
    # We multiply by two factors of the platescale because we expect the
    # sky_sed to be normalized such that calc_adu gives counts per
    # square arc second, and we need to convert to counts per pixel.

    skycounts = (
        sky_sed.calc_adu(hardwarebandpass, phot_params=phot_params)
        * phot_params.platescale
        * phot_params.platescale
    )

    # Calculate the square of the noise due to instrumental effects.
    # Include the readout noise as many times as there are exposures

    noise_instr_sq = calc_instr_noise_sq(phot_params=phot_params)

    # Calculate the square of the noise due to sky background poisson noise
    noise_sky_sq = skycounts / phot_params.gain

    # Discount error in sky measurement for now
    noise_skymeasurement_sq = 0

    total_noise_sq = neff * (noise_sky_sq + noise_instr_sq + noise_skymeasurement_sq)

    return total_noise_sq


def calc_sky_counts_per_pixel_for_m5(m5target, total_bandpass, phot_params, fwhm_eff=0.83):
    """Calculate the skycounts per pixel.

    Calculate the number of sky counts per pixel expected for a given
    value of the 5-sigma limiting magnitude (m5)

    Parameters
    ----------
    m5target : `float`
        The desired value of m5.
    total_bandpass : `Bandpass`
        A bandpass object representing the total throughput of the telescope
        (instrumentation plus atmosphere).
    phot_params : `PhotometricParameters`
        A photometric parameters object containing the photometric response
        information for Rubin.
    fwhm_eff : `float`
        fwhm_eff in arcseconds. Default 0.83

    Returns
    -------
    sky_counts_target : `float`
        The expected number of sky counts per pixel (ADU/pixel).

    Notes
    -----
    The 5-sigma limiting magnitude (m5) for an observation is
    determined by a combination of the telescope and camera parameters
    (such as diameter of the mirrors and the readnoise) together with the
    sky background.
    """

    # instantiate a flat SED
    flat_sed = Sed()
    flat_sed.set_flat_sed()

    # normalize the SED so that it has a magnitude equal to the desired m5
    f_norm = flat_sed.calc_flux_norm(m5target, total_bandpass)
    flat_sed.multiply_flux_norm(f_norm)
    source_counts = flat_sed.calc_adu(total_bandpass, phot_params=phot_params)

    # calculate the effective number of pixels for a double-Gaussian PSF
    neff = calc_neff(fwhm_eff, phot_params.platescale)

    # calculate the square of the noise due to the instrument
    noise_instr_sq = calc_instr_noise_sq(phot_params=phot_params)

    # now solve equation 41 of the SNR document for the neff * sigma_total^2
    # term given snr=5 and counts as calculated above
    # SNR document can be found at
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    n_sigma_sq = (source_counts * source_counts) / 25.0 - source_counts / phot_params.gain

    sky_noise_target = n_sigma_sq / neff - noise_instr_sq
    sky_counts_target = sky_noise_target * phot_params.gain

    # TODO:
    # This method should throw an error if sky_counts_target is negative
    # unfortunately, that currently happens for default values of
    # m5 as taken from arXiv:0805.2366, table 2.  Adding the error
    # should probably wait for a later issue in which we hash out what
    # the units are for all of the parameters stored in PhotometricDefaults.

    return sky_counts_target


def calc_m5(skysed, total_bandpass, hardware, phot_params, fwhm_eff=0.83):
    """Calculate the AB magnitude of a 5-sigma source above sky background source.

    Parameters
    ----------
    skysed : `Sed`
        An SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    total_bandpass : `Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    hardware : `Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.

    Returns
    -------
    mag_5sigma : `float`
        The value of m5 for the given bandpass and sky SED

    Notes
    -----
    The 5-sigma limiting magnitude (m5) for an observation is determined by
    a combination of the telescope and camera parameters (such as diameter
    of the mirrors and the readnoise) together with the sky background. This
    method (calc_m5) calculates the expected m5 value for an observation given
    a sky background Sed and hardware parameters.

    This comes from equation 45 of the SNR document (v1.2, May 2010)
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    """

    # create a flat fnu source
    flatsource = Sed()
    flatsource.set_flat_sed()
    snr = 5.0
    v_n = calc_total_non_source_noise_sq(skysed, hardware, phot_params, fwhm_eff)

    counts_5sigma = (snr**2) / 2.0 / phot_params.gain + numpy.sqrt(
        (snr**4) / 4.0 / phot_params.gain + (snr**2) * v_n
    )

    # renormalize flatsource so that it has the required counts to be a
    # 5-sigma detection given the specified background
    counts_flat = flatsource.calc_adu(total_bandpass, phot_params=phot_params)
    flatsource.multiply_flux_norm(counts_5sigma / counts_flat)

    # Calculate the AB magnitude of this source.
    mag_5sigma = flatsource.calc_mag(total_bandpass)
    return mag_5sigma


def mag_error_from_snr(snr):
    """Convert flux signal to noise ratio to an error in magnitude.

    Parameters
    ----------
    snr : `float`
        The signal to noise ratio (a flux-related measurement).

    Returns
    -------
    mag_error : `float`
        Corresponding error in magnitude.
    """

    # see www.ucolick.org/~bolte/AY257/s_n.pdf section 3.1
    return 2.5 * numpy.log10(1.0 + 1.0 / snr)


def calc_gamma(bandpass, m5, phot_params):
    """Calculate gamma parameter.

    Calculate the gamma parameter used for determining photometric
    signal to noise in equation 5 of the LSST overview paper
    (arXiv:0805.2366)

    Parameters
    ----------
    bandpass : `Bandpass`
        Bandpass for which you desire to calculate the gamma parameter.
    m5 : `float`
        The magnitude of a 5-sigma point source detection.
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.

    Returns
    -------
    gamma : `float`
        The gamma value for this bandpass.
    """
    # This is based on the LSST SNR document (v1.2, May 2010)
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    # as well as equations 4-6 of the overview paper (arXiv:0805.2366)

    # instantiate a flat SED
    flat_sed = Sed()
    flat_sed.set_flat_sed()

    # normalize the SED so that it has a magnitude equal to the desired m5
    f_norm = flat_sed.calc_flux_norm(m5, bandpass)
    flat_sed.multiply_flux_norm(f_norm)
    counts = flat_sed.calc_adu(bandpass, phot_params=phot_params)

    # The expression for gamma below comes from:
    #
    # 1) Take the approximation N^2 = N0^2 + alpha S from footnote 88
    # in the overview paper where N is the noise in flux of a source,
    # N0 is the noise in flux due to sky brightness and instrumentation,
    # S is the number of counts registered from the source and
    # alpha is some constant
    #
    # 2) Divide by S^2 and demand that N/S = 0.2 for a source detected at m5.
    # Solve the resulting equation for alpha in terms of N0 and S5
    # (the number of counts from a source at m5)
    #
    # 3) Substitute this expression for alpha into the equation for (N/S)^2
    # for a general source.  Re-factor the equation so that it looks like
    # equation 5 of the overview paper (note that x = S5/S).
    # This should give you gamma = (N0/S5)^2
    #
    # 4) Solve equation 41 of the SNR document for neff * sigma_total^2 term
    # given snr=5 and counts as calculated above.
    # Note that neff * sigma_total^2 is N0^2 in the equation above
    #
    # This should give you

    gamma = 0.04 - 1.0 / (counts * phot_params.gain)

    return gamma


def calc_snr_m5(magnitude, bandpass, m5, phot_params, gamma=None):
    """Calculate the SNR of a source based on the 5-sigma limit for an
    observation.

    Calculate signal to noise in flux using the model from equation (5)
    of arXiv:0805.2366

    Parameters
    ----------
    magnitude : `float` or `np.ndarray`, (N,)
        Magnitudes of the sources whose signal to noise you are calculating.
    bandpass : `Bandpass`
        The Bandpass in which the magnitude was calculated
        (total instrument + atmosphere).
    m5 : `float`
        The 5-sigma point source limiting magnitude of the exposure.
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    gamma : `float`, opt
        The gamma parameter from equation(5) of arXiv:0805.2366.
        If not provided, this method will calculate it.

    Returns
    -------
    snr : `float` or `np.ndarray`, (N,)
        The SNR of the input magnitude.
    gamma : `float`
        The gamma parameter for the Bandpass.
    """

    if gamma is None:
        gamma = calc_gamma(bandpass, m5, phot_params=phot_params)

    dummy_sed = Sed()
    m5_flux = dummy_sed.flux_from_mag(m5)
    source_flux = dummy_sed.flux_from_mag(magnitude)

    flux_ratio = m5_flux / source_flux

    noise = numpy.sqrt((0.04 - gamma) * flux_ratio + gamma * flux_ratio * flux_ratio)

    return 1.0 / noise, gamma


def calc_mag_error_m5(magnitude, bandpass, m5, phot_params, gamma=None):
    """Calculate magnitude error using the model from equation (5)
    of arXiv:0805.2366

    Parameters
    ----------
    magnitude : `float`
        Magnitude of the source.
    bandpass : `Bandpass`
        The Bandpass in which to calculate the magnitude error
        (total instrument + atmosphere).
    m5 : `float`
        The 5-sigma point source limiting magnitude.
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    gamma : `float`, optional
        The gamma parameter from equation(5) of arXiv:0805.2366.
        If not provided, this method will calculate it.

    Returns
    -------
    mag_error : `float` or `np.ndarray`, (N,)
        The magnitude error of the input magnitude.
    gamma : `float`
        The gamma parameter for the Bandpass.
    """

    snr, gamma = calc_snr_m5(magnitude, bandpass, m5, phot_params, gamma=gamma)

    if phot_params.sigma_sys is not None:
        return (
            numpy.sqrt(numpy.power(mag_error_from_snr(snr), 2) + numpy.power(phot_params.sigma_sys, 2)),
            gamma,
        )
    else:
        return mag_error_from_snr(snr), gamma


def calc_snr_sed(
    source_sed,
    total_bandpass,
    sky_sed,
    hardwarebandpass,
    phot_params,
    fwhm_eff,
    verbose=False,
):
    """Calculate the signal to noise ratio for a source, based on the SED.

    For a given source, sky sed, total bandpass and hardware bandpass,
    as well af fwhm_eff / exptime, calculates the SNR with
    optimal PSF extraction assuming a double-gaussian PSF.

    Parameters
    ----------
    source_sed : `Sed`
        A SED representing the source, normalized such that source_sed.calc_mag
        gives the desired magnitude.
    total_bandpass : `Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    sky_sed : `Sed`
        A SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    hardware : `Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.
    verbose : `bool`
        Flag as to whether to print output about SNR.

    Returns
    -------
    snr : `float`
        Calculated SNR.
    """

    # Calculate the counts from the source.
    sourcecounts = source_sed.calc_adu(total_bandpass, phot_params=phot_params)

    # Calculate the (square of the) noise due to signal poisson noise.
    noise_source_sq = sourcecounts / phot_params.gain

    non_source_noise_sq = calc_total_non_source_noise_sq(sky_sed, hardwarebandpass, phot_params, fwhm_eff)

    # Calculate total noise
    noise = numpy.sqrt(noise_source_sq + non_source_noise_sq)
    # Calculate the signal to noise ratio.
    snr = sourcecounts / noise
    if verbose:
        skycounts = sky_sed.calc_adu(hardwarebandpass, phot_params) * (phot_params.platescale**2)
        noise_sky_sq = skycounts / phot_params.gain
        neff = calc_neff(fwhm_eff, phot_params.platescale)
        noise_instr_sq = calc_instr_noise_sq(phot_params)

        print("For Nexp %.1f of time %.1f: " % (phot_params.nexp, phot_params.exptime))
        print("Counts from source: %.2f  Counts from sky: %.2f" % (sourcecounts, skycounts))
        print("fwhm_eff: %.2f('')  Neff pixels: %.3f(pix)" % (fwhm_eff, neff))
        print(
            "Noise from sky: %.2f Noise from instrument: %.2f"
            % (numpy.sqrt(noise_sky_sq), numpy.sqrt(noise_instr_sq))
        )
        print("Noise from source: %.2f" % (numpy.sqrt(noise_source_sq)))
        print(" Total Signal: %.2f   Total Noise: %.2f    SNR: %.2f" % (sourcecounts, noise, snr))
        # Return the signal to noise value.
    return snr


def calc_mag_error_sed(
    source_sed,
    total_bandpass,
    sky_sed,
    hardware_bandpass,
    phot_params,
    fwhm_eff,
    verbose=False,
):
    """Calculate the magnitudeError for a source, given the bandpass(es)
    and sky SED.

    For a given source, sky sed, total bandpass and hardware bandpass, and
    fwhm_eff / exptime, calculates the SNR with optimal PSF extraction
    assuming a double-gaussian PSF.

    Parameters
    ----------
    source_sed : `Sed`
        A SED representing the source, normalized such that source_sed.calc_mag
        gives the desired magnitude.
    total_bandpass : `Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    sky_sed : `Sed`
        A SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    hardware_bandpass : `Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.
    verbose : `bool`, optional
        Flag as to whether to print output about SNR.

    Returns
    -------
    mag_err : `float`
        Magnitude error in expected magnitude.
    """

    snr = calc_snr_sed(
        source_sed,
        total_bandpass,
        sky_sed,
        hardware_bandpass,
        phot_params,
        fwhm_eff,
        verbose=verbose,
    )

    if phot_params.sigma_sys is not None:
        return numpy.sqrt(numpy.power(mag_error_from_snr(snr), 2) + numpy.power(phot_params.sigma_sys, 2))
    else:
        return mag_error_from_snr(snr)


def calc_astrometric_error(mag, m5, fwhm_geom=0.7, nvisit=1, systematic_floor=10):
    """Calculate an expected astrometric error.

    The astrometric error can be estimated for catalog purposes by using the
    typical FWHM and n_visit = total number of visits, or can be used for a
    single visit by using the actual FWHM and n_visit =1.


    Parameters
    ----------
    mag: `float`
        Magnitude of the source
    m5: `float`
        Point source five sigma limiting magnitude of the image
        (or typical depth per image).
    fwhm_geom: `float`, optional
        The geometric (physical) FWHM of the image, in arcseconds.
    nvisit: `int`, optional
        The number of visits/measurement. Default 1.
        If this is >1, the random error contribution is reduced by
        sqrt(nvisits).
    systematic_floor: `float`, optional
        The systematic noise floor for the astrometric measurements,
        in mas. Default 10mas.

    Returns
    -------
    astrom_err : `float`
        Astrometric error for a given SNR, in mas.
    """
    # The astrometric error can be applied to parallax or proper motion (for n_visit>1).
    # If applying to proper motion, should also divide by the # of years of the survey.
    # This is also referenced in the astroph/0805.2366 paper.
    # D. Monet suggests sqrt(Nvisit/2) for first 3 years, sqrt(N) for longer, in reduction of error
    # because of the astrometric measurement method, the systematic and random error are both reduced.
    # Zeljko says 'be conservative', so removing this reduction for now.
    rgamma = 0.039
    xval = numpy.power(10, 0.4 * (mag - m5))
    # The average fwhm_eff is 0.7" (or 700 mas), but user can specify. Convert to mas.
    seeing = fwhm_geom * 1000.0
    error_rand = seeing * numpy.sqrt((0.04 - rgamma) * xval + rgamma * xval * xval)
    error_rand = error_rand / numpy.sqrt(nvisit)
    # The systematic error floor in astrometry (mas).
    error_sys = systematic_floor
    astrom_error = numpy.sqrt(error_sys * error_sys + error_rand * error_rand)
    return astrom_error



import numpy as np

# Use the main stack to make a rough array.


# Need to put these in sims_utils and remove from MAF and scheduler.


def gnomonic_project_toxy(ra1, dec1, r_acen, deccen):
    """Calculate x/y projection of ra1/dec1 in system with center at r_acen, deccen.
    Input radians. Grabbed from sims_selfcal"""
    # also used in Global Telescope Network website
    cosc = np.sin(deccen) * np.sin(dec1) + np.cos(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)
    x = np.cos(dec1) * np.sin(ra1 - r_acen) / cosc
    y = (np.cos(deccen) * np.sin(dec1) - np.sin(deccen) * np.cos(dec1) * np.cos(ra1 - r_acen)) / cosc
    return x, y


def gnomonic_project_tosky(x, y, r_acen, deccen):
    """Calculate RA/dec on sky of object with x/y and RA/Cen of field of view.
    Returns Ra/dec in radians."""
    denom = np.cos(deccen) - y * np.sin(deccen)
    RA = r_acen + np.arctan2(x, denom)
    dec = np.arctan2(np.sin(deccen) + y * np.cos(deccen), np.sqrt(x * x + denom * denom))
    return RA, dec


if __name__ == "__main__":
    import lsst.sims.utils as simsUtils
    from lsst.obs.lsst import LsstCamMapper
    from lsst.sims.coordUtils import _chipNameFromRaDec

    mapper = LsstCamMapper()
    camera = mapper.camera
    epoch = 2000.0

    ra = 0.0
    dec = 0.0
    rot_sky_pos = 0.0
    mjd = 5300.0

    obs_metadata = simsUtils.ObservationMetaData(
        pointing_ra=np.degrees(ra),
        pointing_dec=np.degrees(dec),
        rot_sky_pos=np.degrees(rot_sky_pos),
        mjd=mjd,
    )

    nside = int(1000)
    # 60k pixels, from 0 to 3.5 degrees
    x_one = np.linspace(-1.75, 1.75, int(nside))

    # make 2-d x,y arrays
    x_two = np.broadcast_to(x_one, (nside, nside))
    y_two = np.broadcast_to(x_one, (nside, nside)).T

    result = np.ones((nside, nside), dtype=bool)
    ra_two, dec_two = gnomonic_project_tosky(np.radians(x_two), np.radians(y_two), ra, dec)
    chip_names = _chipNameFromRaDec(
        ra_two.ravel(),
        dec_two.ravel(),
        epoch=epoch,
        camera=camera,
        obs_metadata=obs_metadata,
    )

    chip_names = chip_names.reshape(nside, nside)
    wavefront_names = [
        name
        for name in np.unique(chip_names[np.where(chip_names != None)])
        if ("SW" in name) | ("R44" in name) | ("R00" in name) | ("R04" in name) | ("R40" in name)
    ]
    # If it's on a waverfront sensor, that's false
    for name in wavefront_names:
        result[np.where(chip_names == name)] = False
    # No chipname, that's a false
    result[np.where(chip_names == None)] = False

    np.savez("fov_map.npz", x=x_one, image=result)



__all__ = ("SFUncertMetric",)

import warnings

import numpy as np
from astropy.stats import mad_std

from rubin_sim.maf.utils import m52snr
from rubin_sim.phot_utils import DustValues

from .base_metric import BaseMetric


class SFUncertMetric(BaseMetric):
    """Structure Function (SF) Uncertainty Metric. Developed on top of LogTGaps

    Adapted from Weixiang Yu & Gordon Richards at:
    https://github.com/RichardsGroup/LSST_SF_Metric/blob/main/notebooks/00_SFErrorMetric.ipynb

    Parameters
    ----------
    mag: `float` (22)
        The magnitude of the fiducial object. Default 22.
    times_col: `str`  ('observationStartMJD')
        Time column name. Defaults to "observationStartMJD".
    all_gaps: `bool` (True)
         Whether to use all gaps (between any two pairs of observations).
         If False, only use consecutive paris. Defaults to True.
    units: `str` ('mag')
        Unit of this metric. Defaults to "mag".
    bins: `object`
        An array of bin edges. Defaults to "np.logspace(0, np.log10(3650), 16)" for a
        total of 15 (final) bins.
    weight: `object
        The weight assigned to each delta_t bin for deriving the final metric.
        Defaults to flat weighting with sum of 1. Should have length 1 less than bins.
    snr_cut : float (5)
        Ignore observations below an SNR limit, default 5.
    dust : `bool` (True)
        Apply dust extinction to the fiducial object magnitude. Default True.
    """

    def __init__(
        self,
        mag=22,
        times_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        all_gaps=True,
        units="mag",
        bins=np.logspace(0, np.log10(3650), 16),
        weight=None,
        metric_name="Structure Function Uncert",
        snr_cut=5,
        filter_col="filter",
        dust=True,
        **kwargs,
    ):
        # Assign metric parameters to instance object
        self.times_col = times_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.all_gaps = all_gaps
        self.bins = bins
        if weight is None:
            # If weight is none, set weight so that sum over bins = 1
            self.weight = np.ones(len(self.bins) - 1)
            self.weight /= self.weight.sum()
        self.metric_name = metric_name
        self.mag = mag
        self.snr_cut = snr_cut
        self.dust = dust

        maps = ["DustMap"]
        super(SFUncertMetric, self).__init__(
            col=[self.times_col, m5_col, filter_col],
            metric_name=self.metric_name,
            units=units,
            maps=maps,
            **kwargs,
        )
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

    def run(self, data_slice, slice_point=None):
        """Code executed at each healpix pixel to compute the metric"""

        df = np.unique(data_slice[self.filter_col])
        if np.size(df) > 1:
            msg = """Running structure function on multiple filters simultaneously.
                     Should probably change your SQL query to limit to a single filter."""
            warnings.warn(msg)
        if self.dust:
            a_x = self.ax1[data_slice[self.filter_col][0]] * slice_point["ebv"]
            extincted_mag = self.mag + a_x
        else:
            extincted_mag = self.mag
        snr = m52snr(extincted_mag, data_slice[self.m5_col])
        bright_enough = np.where(snr > self.snr_cut)[0]

        # If the total number of visits < 2, mask as bad pixel
        if data_slice[bright_enough].size < 2:
            return self.badval

        # sort data by time column
        order = np.argsort(data_slice[self.times_col][bright_enough])
        times = data_slice[self.times_col][bright_enough][order]
        # Using the simple Gaussian approximation for magnitude uncertainty.
        mag_err = 2.5 * np.log10(1.0 + 1.0 / snr[bright_enough][order])

        # check if use all gaps (between any pairs of observations)
        if self.all_gaps:
            # use the vectorized method
            dt_matrix = times.reshape((1, times.size)) - times.reshape((times.size, 1))
            dts = dt_matrix[dt_matrix > 0].flatten().astype(np.float16)
        else:
            dts = np.diff(times)

        # bin delta_t using provided bins; if zero pair found at any delta_t bin,
        # replace 0 with 0.01 to avoid the exploding 1/sqrt(n) term in this metric
        result, bins = np.histogram(dts, self.bins)
        new_result = np.where(result > 0, result, 0.01)

        # compute photometric_error^2 population variance and population mean
        # note that variance is replaced by median_absolute_deviate^2
        # mean is replaced by median in this implementation to make it robust to
        # outliers in simulations (e.g., dcr simulations)
        err_var = mag_err**2
        err_var_mu = np.median(err_var)
        err_var_std = mad_std(err_var)

        # compute SF error
        sf_var_dt = 2 * (err_var_mu + err_var_std / np.sqrt(new_result))
        sf_var_metric = np.sum(sf_var_dt * self.weight)

        return sf_var_metric



import itertools
import unittest
import warnings

import numpy as np
import numpy.lib.recfunctions as rfn

from rubin_sim.maf.slicers import NDSlicer, UniSlicer


def make_data_values(size=100, min=0.0, max=1.0, nd=3, random=-1):
    """Generate a simple array of numbers, evenly arranged between min/max,
    in nd dimensions, but (optional) random order.
    """
    data = []
    for d in range(nd):
        datavalues = np.arange(0, size, dtype="float")
        datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
        datavalues += min
        if random > 0:
            rng = np.random.RandomState(random)
            randorder = rng.rand(size)
            randind = np.argsort(randorder)
            datavalues = datavalues[randind]
        datavalues = np.array(list(zip(datavalues)), dtype=[("testdata" + "%d" % (d), "float")])
        data.append(datavalues)
    data = rfn.merge_arrays(data, flatten=True, usemask=False)
    return data


class TestNDSlicerSetup(unittest.TestCase):
    def setUp(self):
        self.dvmin = 0
        self.dvmax = 1
        nvalues = 1000
        self.nd = 3
        self.dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=608)
        self.dvlist = self.dv.dtype.names

    def test_slicertype(self):
        """Test instantiation of slicer sets slicer type as expected."""
        testslicer = NDSlicer(self.dvlist)
        self.assertEqual(testslicer.slicer_name, testslicer.__class__.__name__)
        self.assertEqual(testslicer.slicer_name, "NDSlicer")

    def test_setup_slicer_bins(self):
        """Test setting up slicer using defined bins."""
        # Used right bins?
        bins = np.arange(self.dvmin, self.dvmax, 0.1)
        binlist = []
        for d in range(self.nd):
            binlist.append(bins)
        testslicer = NDSlicer(self.dvlist, bins_list=binlist)
        testslicer.setup_slicer(self.dv)
        for d in range(self.nd):
            np.testing.assert_equal(testslicer.bins[d], bins)
        self.assertEqual(testslicer.nslice, (len(bins) - 1) ** self.nd)

    def test_setup_slicer_nbins(self):
        """Test setting up slicer using nbins."""
        for nvalues in (100, 1000):
            for nbins in (5, 25, 74):
                dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=-1)
                # Right number of bins?
                # expect one more 'bin' to accomodate last right edge,
                # but nbins accounts for this
                testslicer = NDSlicer(self.dvlist, bins_list=nbins)
                testslicer.setup_slicer(dv)
                self.assertEqual(testslicer.nslice, nbins**self.nd)
                # Bins of the right size?
                for i in range(self.nd):
                    bindiff = np.diff(testslicer.bins[i])
                    expectedbindiff = (self.dvmax - self.dvmin) / float(nbins)
                    np.testing.assert_allclose(bindiff, expectedbindiff)
                # Can we use a list of nbins too and get the
                # right number of bins?
                nbins_list = []
                expectednbins = 1
                for d in range(self.nd):
                    nbins_list.append(nbins + d)
                    expectednbins *= nbins + d
                testslicer = NDSlicer(self.dvlist, bins_list=nbins_list)
                testslicer.setup_slicer(dv)
                self.assertEqual(testslicer.nslice, expectednbins)

    def test_setup_slicer_nbins_zeros(self):
        """Test handling case of data being single values."""
        dv = make_data_values(100, 0, 0, self.nd, random=-1)
        nbins = 10
        testslicer = NDSlicer(self.dvlist, bins_list=nbins)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            testslicer.setup_slicer(dv)
            self.assertIn("creasing bin_max", str(w[-1].message))
        expectednbins = nbins**self.nd
        self.assertEqual(testslicer.nslice, expectednbins)

    def test_setup_slicer_equivalent(self):
        """Test setting up slicer using defined bins and nbins is
        equal where expected.
        """
        for nbins in (20, 105):
            testslicer = NDSlicer(self.dvlist, bins_list=nbins)
            bins = make_data_values(nbins + 1, self.dvmin, self.dvmax, self.nd, random=-1)
            bins_list = []
            for i in bins.dtype.names:
                bins_list.append(bins[i])
            for nvalues in (100, 10000):
                dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=64432)
                testslicer.setup_slicer(dv)
                for i in range(self.nd):
                    np.testing.assert_allclose(testslicer.bins[i], bins_list[i])


class TestNDSlicerEqual(unittest.TestCase):
    def setUp(self):
        self.dvmin = 0
        self.dvmax = 1
        nvalues = 1000
        self.nd = 3
        self.dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=20367)
        self.dvlist = self.dv.dtype.names
        self.testslicer = NDSlicer(self.dvlist, bins_list=100)
        self.testslicer.setup_slicer(self.dv)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_equivalence(self):
        """Test equals method."""
        # Note that two ND slicers will be considered equal if
        # they are both the same kind of slicer AND have the same bins in
        # all dimensions.
        # Set up another slicer to match
        # (same bins, although not the same data).
        dv2 = make_data_values(100, self.dvmin, self.dvmax, self.nd, random=10029)
        dvlist = dv2.dtype.names
        testslicer2 = NDSlicer(slice_col_list=dvlist, bins_list=self.testslicer.bins)
        testslicer2.setup_slicer(dv2)
        self.assertEqual(self.testslicer, testslicer2)
        # Set up another slicer that should not match (different bins)
        dv2 = make_data_values(1000, self.dvmin + 1, self.dvmax + 1, self.nd, random=209837)
        testslicer2 = NDSlicer(slice_col_list=dvlist, bins_list=100)
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)
        # Set up another slicer that should not match (different dimensions)
        dv2 = make_data_values(1000, self.dvmin, self.dvmax, self.nd - 1, random=50623)
        testslicer2 = NDSlicer(dv2.dtype.names, bins_list=100)
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)
        # Set up a different kind of slicer that should not match.
        testslicer2 = UniSlicer()
        dv2 = make_data_values(100, 0, 1, random=22310098)
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)


class TestNDSlicerIteration(unittest.TestCase):
    def setUp(self):
        self.dvmin = 0
        self.dvmax = 1
        nvalues = 1000
        self.nd = 3
        self.dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=11081)
        self.dvlist = self.dv.dtype.names
        nvalues = 1000
        bins = np.arange(self.dvmin, self.dvmax, 0.1)
        bins_list = []
        self.iterlist = []
        for i in range(self.nd):
            bins_list.append(bins)
            # (remember iteration doesn't use the very last bin in 'bins')
            self.iterlist.append(bins[:-1])
        dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=17)
        self.testslicer = NDSlicer(self.dvlist, bins_list=bins_list)
        self.testslicer.setup_slicer(dv)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_iteration(self):
        """Test iteration."""
        for s, ib in zip(self.testslicer, itertools.product(*self.iterlist)):
            self.assertEqual(s["slice_point"]["bin_left"], ib)

    def test_get_item(self):
        """Test getting indexed binpoint."""
        for i, s in enumerate(self.testslicer):
            self.assertEqual(
                self.testslicer[i]["slice_point"]["bin_left"],
                s["slice_point"]["bin_left"],
            )
        self.assertEqual(self.testslicer[0]["slice_point"]["bin_left"], (0.0, 0.0, 0.0))


class TestNDSlicerSlicing(unittest.TestCase):
    def setUp(self):
        self.dvmin = 0
        self.dvmax = 1
        nvalues = 1000
        self.nd = 3
        self.dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=173)
        self.dvlist = self.dv.dtype.names
        self.testslicer = NDSlicer(self.dvlist)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicing(self):
        """Test slicing."""
        # Test get error if try to slice before setup.
        self.assertRaises(NotImplementedError, self.testslicer._slice_sim_data, 0)
        nbins = 10
        bin_size = (self.dvmax - self.dvmin) / (float(nbins))
        self.testslicer = NDSlicer(self.dvlist, bins_list=nbins)
        for nvalues in (1000, 10000):
            dv = make_data_values(nvalues, self.dvmin, self.dvmax, self.nd, random=1735)
            self.testslicer.setup_slicer(dv)
            sum = 0
            for i, s in enumerate(self.testslicer):
                idxs = s["idxs"]
                data_slice = dv[idxs]
                sum += len(idxs)
                if len(data_slice) > 0:
                    for i, dvname, b in zip(list(range(self.nd)), self.dvlist, s["slice_point"]["bin_left"]):
                        self.assertGreaterEqual((data_slice[dvname].min() - b), 0)
                    if i < self.testslicer.nslice - 1:
                        self.assertLessEqual((data_slice[dvname].max() - b), bin_size)
                    else:
                        self.assertAlmostEqual((data_slice[dvname].max() - b), bin_size)
                    self.assertEqual(len(data_slice), nvalues / float(nbins))
            # and check that every data value was assigned somewhere.
            self.assertEqual(sum, nvalues)


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings

import numpy as np

from rubin_sim.maf.slicers import MovieSlicer, UniSlicer


def make_times(size=100, min=0.0, max=10.0, random=-1):
    """Generate a simple array of numbers, evenly arranged between min/max."""
    datavalues = np.arange(0, size, dtype="float")
    datavalues *= (float(max) - float(min)) / (datavalues.max() - datavalues.min())
    datavalues += min
    if random > 0:
        rng = np.random.RandomState(random)
        randorder = rng.rand(size)
        randind = np.argsort(randorder)
        datavalues = datavalues[randind]
    datavalues = np.array(list(zip(datavalues)), dtype=[("times", "float")])
    return datavalues


class TestMovieSlicerSetup(unittest.TestCase):
    def setUp(self):
        self.testslicer = MovieSlicer(slice_col_name="times", cumulative=False, force_no_ffmpeg=True)

    def tearDown(self):
        del self.testslicer
        self.testslicer = None

    def test_slicertype(self):
        """Test instantiation of slicer sets slicer type as expected."""
        self.assertEqual(self.testslicer.slicer_name, self.testslicer.__class__.__name__)
        self.assertEqual(self.testslicer.slicer_name, "MovieSlicer")

    def test_setup_slicer_bins(self):
        """Test setting up slicer using defined bins."""
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        bins = np.arange(dvmin, dvmax, 0.1)
        dv = make_times(nvalues, dvmin, dvmax, random=987)
        # Used right bins?
        self.testslicer = MovieSlicer(
            slice_col_name="times", bins=bins, cumulative=False, force_no_ffmpeg=True
        )
        self.testslicer.setup_slicer(dv)
        np.testing.assert_equal(self.testslicer.bins, bins)
        self.assertEqual(self.testslicer.nslice, len(bins) - 1)

    def test_setup_slicer_binsize(self):
        """Test setting up slicer using bin_size."""
        dvmin = 0
        dvmax = 1
        dv = make_times(1000, dvmin, dvmax, random=1992)
        bin_size = 0.1
        for cumulative in [True, False]:
            self.testslicer = MovieSlicer(
                slice_col_name="times",
                bin_size=bin_size,
                cumulative=cumulative,
                force_no_ffmpeg=True,
            )
            self.testslicer.setup_slicer(dv)
            # Bins of the right size?
            bindiff = np.diff(self.testslicer.bins)
            self.assertAlmostEqual(bindiff.max(), bin_size)
            self.assertAlmostEqual(bindiff.min(), bin_size)
        # Test that warning works.
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.testslicer = MovieSlicer(
                slice_col_name="times",
                bins=200,
                bin_size=bin_size,
                cumulative=False,
                force_no_ffmpeg=True,
            )
            self.testslicer.setup_slicer(dv)
            # Verify some things
            self.assertIn("bin_size", str(w[-1].message))

    def test_setup_slicer_nbins_zeros(self):
        """Test what happens if give slicer test data that is all
        single-value."""
        dv = np.zeros(100, float)
        dv = np.array(list(zip(dv)), dtype=[("times", "float")])
        nbins = 10
        self.testslicer = MovieSlicer(
            slice_col_name="times", bins=nbins, cumulative=False, force_no_ffmpeg=True
        )
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            self.testslicer.setup_slicer(dv)
            self.assertIn("creasing bin_max", str(w[-1].message))
        self.assertEqual(self.testslicer.nslice, nbins)

    def test_setup_slicer_limits(self):
        """Test setting up slicer using bin_min/Max."""
        bin_min = 0
        bin_max = 1
        nbins = 10
        dvmin = -0.5
        dvmax = 1.5
        dv = make_times(1000, dvmin, dvmax, random=1772)
        self.testslicer = MovieSlicer(
            slice_col_name="times",
            bin_min=bin_min,
            bin_max=bin_max,
            bins=nbins,
            cumulative=False,
            force_no_ffmpeg=True,
        )
        self.testslicer.setup_slicer(dv)
        self.assertAlmostEqual(self.testslicer.bins.min(), bin_min)
        self.assertAlmostEqual(self.testslicer.bins.max(), bin_max)

    def test_indexing(self):
        """Test iteration and indexing."""
        dvmin = 0
        dvmax = 1
        bins = np.arange(dvmin, dvmax + 0.05, 0.05)
        self.testslicer = MovieSlicer(
            slice_col_name="times", bins=bins, cumulative=False, force_no_ffmpeg=True
        )
        dv = make_times(1000, dvmin, dvmax, random=908223)
        self.testslicer.setup_slicer(dv)
        for i, (s, b) in enumerate(zip(self.testslicer, bins)):
            self.assertEqual(s["slice_point"]["sid"], i)
            self.assertEqual(s["slice_point"]["bin_left"], b)
            self.assertLessEqual(s["slice_point"]["bin_right"], bins[i + 1])
        for i in [0, len(self.testslicer) // 2, len(self.testslicer) - 1]:
            self.assertEqual(self.testslicer[i]["slice_point"]["sid"], i)
            self.assertEqual(self.testslicer[i]["slice_point"]["bin_left"], bins[i])

    def test_equivalence(self):
        """Test equals method."""
        # Note that two Movie slicers will be considered equal
        # if they are both the same kind of
        # slicer AND have the same bins.
        # Set up self..
        dvmin = 0
        dvmax = 1
        nvalues = 1000
        bins = np.arange(dvmin, dvmax, 0.01)
        dv = make_times(nvalues, dvmin, dvmax, random=72031)
        self.testslicer = MovieSlicer(
            slice_col_name="times", bins=bins, cumulative=False, force_no_ffmpeg=True
        )
        self.testslicer.setup_slicer(dv)
        # Set up another slicer to match
        # (same bins, although not the same data).
        dv2 = make_times(nvalues + 100, dvmin, dvmax, random=56221)
        testslicer2 = MovieSlicer(slice_col_name="times", bins=bins, cumulative=False, force_no_ffmpeg=True)
        testslicer2.setup_slicer(dv2)
        self.assertEqual(self.testslicer, testslicer2)
        # Set up another slicer that should not match (different bins)
        dv2 = make_times(nvalues, dvmin + 1, dvmax + 1, random=542093)
        testslicer2 = MovieSlicer(
            slice_col_name="times",
            bins=len(bins),
            cumulative=False,
            force_no_ffmpeg=True,
        )
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)
        # Set up a different kind of slicer that should not match.
        dv2 = make_times(100, 0, 1, random=16)
        testslicer2 = UniSlicer()
        testslicer2.setup_slicer(dv2)
        self.assertNotEqual(self.testslicer, testslicer2)

    def test_slicing(self):
        """Test slicing."""
        dvmin = 0
        dvmax = 1
        nbins = 100
        # Test that testbinner raises appropriate error before it's set up
        # (first time)
        self.assertRaises(NotImplementedError, self.testslicer._slice_sim_data, 0)
        for nvalues in (100, 1000):
            dv = make_times(nvalues, dvmin, dvmax, random=82)
            # Test differential case.
            self.testslicer = MovieSlicer(
                slice_col_name="times",
                bins=nbins,
                cumulative=False,
                force_no_ffmpeg=True,
            )
            self.testslicer.setup_slicer(dv)
            sum = 0
            for i, s in enumerate(self.testslicer):
                idxs = s["idxs"]
                data_slice = dv["times"][idxs]
                sum += len(idxs)
                if len(data_slice) > 0:
                    self.assertEqual(len(data_slice), nvalues / float(nbins))
                else:
                    raise ValueError("Data in test case expected to always be > 0 len after slicing")
            self.assertEqual(sum, nvalues)
            # And cumulative case.
            self.testslicer = MovieSlicer(
                slice_col_name="times",
                bins=nbins,
                cumulative=True,
                force_no_ffmpeg=True,
            )
            self.testslicer.setup_slicer(dv)
            for i, s in enumerate(self.testslicer):
                idxs = s["idxs"]
                data_slice = dv["times"][idxs]
                self.assertGreater(len(data_slice), 0)


if __name__ == "__main__":
    unittest.main()
